/*
 * Copyright (C) 2006 Texas Instruments, Incorporated
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "gstomxaudiodec.h"

#include <gst/gst.h>
#include <config.h>

/* For memcpy. */
#include <string.h>

GST_DEBUG_CATEGORY_STATIC (gomx_audiodec_debug);
#define GST_CAT_DEFAULT gomx_audiodec_debug

#define DEFAULT_NUM_INPUT_BUFFERS 1
#define DEFAULT_NUM_OUTPUT_BUFFERS 1

/* AAC Defaults */

#define DEFAULT_AAC_PROFILE 1
#define DEFAULT_SBR 1
#define DEFAULT_RAW 0
#define DEFAULT_PARAMETRIC_STEREO 0
#define DEFAULT_BIT_OUTPUT 0
#define DEFAULT_MIME 0

typedef enum _GstOmxAudioDecElementId GstOmxAudioDecElementId;

enum _GstOmxAudioDecElementId
{
	GST_ELEMENT_NULL,
	GST_ELEMENT_AMR,
	GST_ELEMENT_AAC,
	GST_ELEMENT_MP3,
	GST_ELEMENT_WBAMR,
	GST_ELEMENT_PCM
};

#define GST_TYPE_AAC_PROFILE (gst_gomx_audiodec_aac_profile_get_type())

static GType
gst_gomx_audiodec_aac_profile_get_type ()
{
	static GType gomx_audiodec_aac_profile_type = 0;
	static GEnumValue gomx_audiodec_aac_profile[] =
	{
		{0, "0", "EProfileMain"},
		{1, "1", "EProfileLC"},
		{2, "2", "EProfileSSR"},
		{3, "3", "EProfileLTP"},
		{0, NULL, NULL},
	};
	if (!gomx_audiodec_aac_profile_type)
	{
		gomx_audiodec_aac_profile_type = g_enum_register_static ("GstGOmxAudioDecAacProfile", gomx_audiodec_aac_profile);
	}

	return gomx_audiodec_aac_profile_type;
}

enum
{
	LAST_SIGNAL
};

enum
{
	ARG_0,
	ARG_SILENT,
	ARG_NUM_INPUT_BUFFERS,
	ARG_NUM_OUTPUT_BUFFERS,
	/*AAC*/
	ARG_EPROFILE,
	ARG_SBR,
	ARG_RAW,
	ARG_PARAMETRIC_STEREO,
	ARG_BIT_OUTPUT,
	/*AMR and WBAMR*/
	ARG_MIME
};

typedef struct
{
	GstGOmxAudioDecElementInfo *info;

	GstCaps *srccaps;
	GstCaps *sinkcaps;
} GstGOmxAudioDecClassParams;

GstGOmxAudioDecElementInfo comps[] =
{
	{
		.id = GST_ELEMENT_AMR,
		.omx_component = "OMX_AmrDecoder",
		.name = "dec_amr",
		.capitalized_name = "DecAmr",
		.longname = "AMR",
		.category = "Codec/Decoder/Audio"
	},
	{
		.id = GST_ELEMENT_AAC,
		.omx_component = "OMX_AacDecoder",
		.name = "dec_aac",
		.capitalized_name = "DecAac",
		.longname = "AAC",
		.category = "Codec/Decoder/Audio"
	},
	{
		.id = GST_ELEMENT_MP3,
		.omx_component ="OMX_Mp3Decoder",
		.name = "dec_mp3",
		.capitalized_name = "DecMp3",
		.longname = "MP3",
		.category = "Codec/Decoder/Audio"
	},
	{
		.id = GST_ELEMENT_WBAMR,
		.omx_component = "OMX_WBAmrDecoder",
		.name = "dec_wbamr",
		.capitalized_name = "DecWbAmr",
		.longname = "WideBand AMR",
		.category = "Codec/Decoder/Audio"
	},
	{
		.id = GST_ELEMENT_PCM,
		.omx_component = "OMX_PcmDecoder",
		.name = "dec_pcm",
		.capitalized_name = "DecPcm",
		.longname = "Pulse Code Modulation",
		.category = "Codec/Decoder/Audio",
	},
	{
		.id = GST_ELEMENT_NULL,
		.omx_component = NULL,
		.name = NULL,
		.capitalized_name = NULL,
		.longname = NULL,
		.category = NULL,
	}
};

static GHashTable *enc_global_plugins;

static GstStaticPadTemplate src_factory =
GST_STATIC_PAD_TEMPLATE ("src",
						 GST_PAD_SRC,
						 GST_PAD_ALWAYS,
						 GST_STATIC_CAPS ("audio/x-raw-int, "
										  "width = (int) [8, 32], "
										  "height = (int) [8, 32], "
										  "signed = (boolean) true, "
										  "rate = (int) [8000, 48000], "
										  "channels = (int) [1, 8]")
						);

/* A number of functon prototypes are given so we can refer to them later. */
static void gst_gomx_audiodec_class_init (GstGOmxAudioDecClass *klass);
static void gst_gomx_audiodec_base_init (GstGOmxAudioDecClass *klass);
static void gst_gomx_audiodec_init (GstGOmxAudioDec *audiodec);
static void gst_gomx_audiodec_dispose (GObject *object);

static GstElementClass *parent_class = NULL;

static void gst_gomx_audiodec_set_property (GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec);
static void gst_gomx_audiodec_get_property (GObject *object, guint prop_id, GValue *value, GParamSpec *pspec);

/* static gboolean gst_gomx_audiodec_src_setcaps (GstPad *pad, GstCaps *caps); */
static gboolean gst_gomx_audiodec_sink_setcaps (GstPad *pad, GstCaps *caps);
static GstFlowReturn gst_gomx_audiodec_chain (GstPad *pad, GstBuffer *buf);

/* */

static void
got_input_cb (GOmxCore *core,
			  OMX_BUFFERHEADERTYPE *omx_buffer)
{
	GstGOmxAudioDec *self;

	self = core->client_data;

	GST_DEBUG_OBJECT (self, "got_input_cb\n");

	core->omx_buffer = omx_buffer;
	g_omx_sem_up (self->input_ready);
	g_omx_sem_down (self->input_filled);
	core->omx_buffer = NULL;
}

static void
got_output_cb (GOmxCore *core,
			   OMX_BUFFERHEADERTYPE *omx_buffer)
{
	GstBuffer *new_buf;
	GstGOmxAudioDec *self;

	self = core->client_data;

	GST_DEBUG_OBJECT (self, "got_output_cb\n");

	new_buf = gst_buffer_new_and_alloc (omx_buffer->nFilledLen);
	/** @todo Get rid of this memcpy. */
	memcpy (GST_BUFFER_DATA (new_buf), omx_buffer->pBuffer, omx_buffer->nFilledLen);

	gst_buffer_set_caps (new_buf, GST_PAD_CAPS (self->srcpad));

	gst_pad_push (self->srcpad, new_buf);
}

/* */

static gboolean
gst_gomx_audiodec_sink_event (GstPad *pad,
							  GstEvent *event)
{
	GstGOmxAudioDec *self;

	self = GST_GOMX_AUDIODEC (GST_OBJECT_PARENT (pad));

	switch (GST_EVENT_TYPE (event))
	{
		case GST_EVENT_EOS:

			GST_DEBUG_OBJECT (self, "eos");

			g_omx_core_set_eos (self->gomx);

			/* We don't want input_cb to be blocked. */
			g_omx_sem_up (self->input_filled);

			break;

		default:
			break;
	}

	return gst_pad_event_default (pad, event);
}

static GstStateChangeReturn
gst_gomx_audiodec_change_state (GstElement *element,
								GstStateChange transition)
{
	GstStateChangeReturn ret = GST_STATE_CHANGE_SUCCESS;

	GstGOmxAudioDec *self = GST_GOMX_AUDIODEC (element);

	switch (transition)
	{
		case GST_STATE_CHANGE_NULL_TO_READY:

			g_omx_core_init (self->gomx, self->info->omx_component);
			break;

		default:
			break;
	}

	ret = GST_ELEMENT_CLASS (parent_class)->change_state (element, transition);

	if (ret == GST_STATE_CHANGE_FAILURE)
		return ret;

	switch (transition)
	{
		case GST_STATE_CHANGE_PAUSED_TO_READY:
			g_omx_core_finish (self->gomx);
			break;

		case GST_STATE_CHANGE_READY_TO_NULL:
			g_omx_core_deinit (self->gomx);
			break;

		default:
			break;
	}

	return ret;
}

static void
gst_gomx_audiodec_dispose (GObject *obj)
{
	GstGOmxAudioDec *self = (GstGOmxAudioDec *) obj;

	if (self->dispose_has_run)
	{
		/* If dispose did already run, return. */
		return;
	}
	/* Make sure dispose does not run twice. */
	self->dispose_has_run = TRUE;

	g_omx_core_free (self->gomx);
	g_hash_table_destroy (enc_global_plugins);
	g_omx_sem_free (self->input_filled);
	g_omx_sem_free (self->input_ready);

	G_OBJECT_CLASS (parent_class)->dispose (obj);
}


static GstCaps *
generate_sink_template (GstOmxAudioDecElementId id)
{
	GstCaps *caps = NULL;

	switch (id)
	{
		case GST_ELEMENT_AMR:
			{
				caps = gst_caps_new_simple ("audio/AMR",
											"rate", GST_TYPE_INT_RANGE, 8000, 48000,
											"channels", GST_TYPE_INT_RANGE, 1, 8,
											NULL);
				break;
			}
		case GST_ELEMENT_AAC:
			{
				caps = gst_caps_new_simple ("audio/mpeg",
											"mpegversion",G_TYPE_INT,4,
											"rate", GST_TYPE_INT_RANGE, 8000, 96000,
											"channels", GST_TYPE_INT_RANGE, 1, 2,
											NULL);
				break;
			}
		case GST_ELEMENT_MP3:
			{
				caps = gst_caps_new_simple ("audio/mpeg",
											"mpegversion",G_TYPE_INT,1,
											"rate", GST_TYPE_INT_RANGE, 8000, 96000,
											"channels", GST_TYPE_INT_RANGE, 1, 2,
											NULL);
				break;
			}
		case GST_ELEMENT_WBAMR:
			{
				caps = gst_caps_new_simple ("audio/AMR-WB",
											"rate", G_TYPE_INT, 16000,
											"channels", G_TYPE_INT, 1,
											NULL);
				break;
			}
		case GST_ELEMENT_PCM:
			{
				caps = gst_caps_new_simple ("audio/x-raw-int",
											"rate", G_TYPE_INT, 8000,
											"channels", GST_TYPE_INT_RANGE, 1,2,
											NULL);
				break;
			}
	}

	return caps;
}

static void
gst_gomx_audiodec_base_init (GstGOmxAudioDecClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
	GstElementClass *element_class = GST_ELEMENT_CLASS (klass);
	GstGOmxAudioDecClassParams *params;
	GstElementDetails details;

	params = g_hash_table_lookup (enc_global_plugins,
								  GINT_TO_POINTER (G_OBJECT_CLASS_TYPE (gobject_class)));

	if (!params)
	{
		params = g_hash_table_lookup (enc_global_plugins, GINT_TO_POINTER (0));
	}

	g_assert (params);

	details.longname = g_strdup_printf ("OpenMAX %s decoder",
										params->info->longname);
	details.klass = params->info->category;
	details.description = g_strdup_printf ("Decodes %s audio with OpenMAX",
										   params->info->longname);
	details.author = "Texas Instruments";
	gst_element_class_set_details (element_class, &details);
	g_free (details.longname);
	g_free (details.description);

	{
		GstPadTemplate *sinktempl;

		sinktempl = gst_pad_template_new ("sink", GST_PAD_SINK,
										  GST_PAD_ALWAYS, params->sinkcaps);

		gst_element_class_add_pad_template (element_class, sinktempl);
	}

	gst_element_class_add_pad_template (element_class,
										gst_static_pad_template_get (&src_factory));

	klass->info = params->info;
}

static void
gst_gomx_audiodec_class_init (GstGOmxAudioDecClass *klass)
{
	GObjectClass *gobject_class;
	GstElementClass *gstelement_class;

	/* gobject_class = G_OBJECT_CLASS (klass); */
	gobject_class = (GObjectClass *) klass;
	gstelement_class = (GstElementClass *) klass;

	parent_class = g_type_class_ref (GST_TYPE_ELEMENT);

	gobject_class->set_property = gst_gomx_audiodec_set_property;
	gobject_class->get_property = gst_gomx_audiodec_get_property;

	g_object_class_install_property (gobject_class, ARG_SILENT,
									 g_param_spec_boolean ("silent", "Silent",
														   "Produce verbose output ?",
														   FALSE, G_PARAM_READWRITE));

	g_object_class_install_property (gobject_class, ARG_NUM_INPUT_BUFFERS,
									 g_param_spec_uint ("input-buffers", "Input buffers",
														"The number of input buffers",
														0, G_MAXUINT, DEFAULT_NUM_INPUT_BUFFERS, G_PARAM_READWRITE));

	g_object_class_install_property (gobject_class, ARG_NUM_OUTPUT_BUFFERS,
									 g_param_spec_uint ("output-buffers", "Output buffers",
														"The number of output buffers",
														0, G_MAXUINT, DEFAULT_NUM_OUTPUT_BUFFERS, G_PARAM_READWRITE));

	if (klass->info->id == GST_ELEMENT_AMR || klass->info->id == GST_ELEMENT_WBAMR)
	{
		g_object_class_install_property (gobject_class, ARG_MIME,
										 g_param_spec_uint ("mime", "MIME",
															"Enable using MIME headers",
															0, 1, DEFAULT_MIME, G_PARAM_READWRITE));
	}

	if (klass->info->id == GST_ELEMENT_AAC)
	{
		g_object_class_install_property (gobject_class, ARG_EPROFILE,
										 g_param_spec_enum ("profile", "Eprofile",
															"The Eprofile",
															GST_TYPE_AAC_PROFILE, DEFAULT_AAC_PROFILE, G_PARAM_READWRITE));

		g_object_class_install_property (gobject_class, ARG_SBR,
										 g_param_spec_uint ("SBR", "SBR",
															"Used to downsample Fs in DASF",
															0, 1, DEFAULT_SBR,  G_PARAM_READWRITE));

		g_object_class_install_property (gobject_class, ARG_RAW,
										 g_param_spec_uint ("RAW", "RAW",
															"RAW format",
															0, 1, DEFAULT_RAW, G_PARAM_READWRITE));

		g_object_class_install_property (gobject_class, ARG_PARAMETRIC_STEREO,
										 g_param_spec_uint ("parametric-stereo", "Parametric Stereo",
															"Parametric Stereo configuration",
															0, 1, DEFAULT_PARAMETRIC_STEREO, G_PARAM_READWRITE));
	}

	if (klass->info->id == GST_ELEMENT_AAC || klass->info->id == GST_ELEMENT_MP3)
	{
		g_object_class_install_property (gobject_class, ARG_BIT_OUTPUT,
										 g_param_spec_uint ("bit-output", "Bit-output",
															"Setting profiles",
															0, 1, DEFAULT_BIT_OUTPUT, G_PARAM_READWRITE));
	}

	gobject_class->dispose = gst_gomx_audiodec_dispose;
	gstelement_class->change_state = gst_gomx_audiodec_change_state;
}

static void
gst_gomx_audiodec_init (GstGOmxAudioDec *self)
{
	GstElementClass *klass = GST_ELEMENT_GET_CLASS (self);
	GstGOmxAudioDecClass *oclass = (GstGOmxAudioDecClass *) (G_OBJECT_GET_CLASS (self));

	GST_LOG_OBJECT (self, "init");

	self->silent = FALSE;
	self->info = oclass->info;
	
	self->sinkpad =
		gst_pad_new_from_template (gst_element_class_get_pad_template (klass, "sink"), "sink");

	gst_pad_set_setcaps_function (self->sinkpad, gst_gomx_audiodec_sink_setcaps);
	/* gst_pad_set_setcaps_function (self->sinkpad, gst_pad_proxy_setcaps); */
	gst_pad_set_chain_function (self->sinkpad, gst_gomx_audiodec_chain);
	gst_pad_set_event_function (self->sinkpad, gst_gomx_audiodec_sink_event);
	gst_element_add_pad (GST_ELEMENT (self), self->sinkpad);

	self->srcpad =
		gst_pad_new_from_template (gst_element_class_get_pad_template (klass, "src"), "src");

	/* gst_pad_set_setcaps_function (self->srcpad, gst_gomx_audiodec_src_setcaps); */
	/* gst_pad_set_setcaps_function (self->srcpad, gst_pad_proxy_setcaps); */
	/* gst_pad_use_fixed_caps (self->srcpad); */
	gst_pad_use_fixed_caps (self->sinkpad);
	gst_element_add_pad (GST_ELEMENT (self), self->srcpad);
	self->silent = FALSE;

	{
		GOmxCore *gomx;

		self->gomx = gomx = g_omx_core_new ();
		self->gomx->client_data = self;
		g_omx_core_set_id (gomx, self->info->name);

		gomx->num_input_buffers = DEFAULT_NUM_INPUT_BUFFERS;
		gomx->num_output_buffers = DEFAULT_NUM_OUTPUT_BUFFERS;

		if (self->info->id == GST_ELEMENT_AAC)
		{

			g_omx_core_set_uint_param (gomx, "profile", DEFAULT_AAC_PROFILE);
			g_omx_core_set_uint_param (gomx, "sbr", DEFAULT_SBR);
			g_omx_core_set_uint_param (gomx, "raw",  DEFAULT_RAW);
			g_omx_core_set_uint_param (gomx, "parametric_stereo",  DEFAULT_PARAMETRIC_STEREO);
			g_omx_core_set_uint_param (gomx, "bit_output",  DEFAULT_BIT_OUTPUT);
		
		}

		if (self->info->id == GST_ELEMENT_MP3)
		{
			g_omx_core_set_uint_param (gomx, "bit_output", DEFAULT_BIT_OUTPUT);
		}
	}

	self->input_ready = g_omx_sem_new ();
	self->input_filled = g_omx_sem_new ();

	g_omx_sem_set_name (self->input_ready, "input ready");
	g_omx_sem_set_name (self->input_filled, "input filled");
}

static void
gst_gomx_audiodec_set_property (GObject *object,
								guint prop_id,
								const GValue *value,
								GParamSpec *pspec)
{
	GstGOmxAudioDec *self = GST_GOMX_AUDIODEC (object);

	switch (prop_id)
	{
		case ARG_SILENT:
			self->silent = g_value_get_boolean (value);
			break;
		case ARG_NUM_INPUT_BUFFERS:
			self->gomx->num_input_buffers = g_value_get_uint (value);
			break;
		case ARG_NUM_OUTPUT_BUFFERS:
			self->gomx->num_output_buffers = g_value_get_uint (value);
			break;
		case ARG_EPROFILE:
			g_omx_core_set_uint_param (self->gomx, "profile", g_value_get_enum (value));
			break;
		case ARG_SBR:
			g_omx_core_set_uint_param (self->gomx, "sbr", g_value_get_uint (value));
			break;
		case ARG_RAW:
			g_omx_core_set_uint_param (self->gomx, "raw", g_value_get_uint (value));
			break;
		case ARG_PARAMETRIC_STEREO:
			g_omx_core_set_uint_param (self->gomx, "parametric_stereo", g_value_get_uint (value));
			break;
		case ARG_BIT_OUTPUT:
			g_omx_core_set_uint_param (self->gomx, "bit_output", g_value_get_uint (value));
			break;
		case ARG_MIME:
			g_omx_core_set_uint_param (self->gomx, "mime", g_value_get_uint (value));
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
gst_gomx_audiodec_get_property (GObject *object,
								guint prop_id,
								GValue *value,
								GParamSpec *pspec)
{
	GstGOmxAudioDec *self = GST_GOMX_AUDIODEC (object);

	switch (prop_id)
	{
		case ARG_SILENT:
			g_value_set_boolean (value, self->silent);
			break;
		case ARG_NUM_INPUT_BUFFERS:
			g_value_set_uint (value, self->gomx->num_input_buffers);
			break;
		case ARG_NUM_OUTPUT_BUFFERS:
			g_value_set_uint (value, self->gomx->num_output_buffers);
			break;
		case ARG_EPROFILE:
			g_value_set_enum (value, g_omx_core_get_uint_param (self->gomx, "profile"));
			break;
		case ARG_SBR:
			g_value_set_uint (value, g_omx_core_get_uint_param (self->gomx, "sbr"));
			break;
		case ARG_RAW:
			g_value_set_uint (value, g_omx_core_get_uint_param (self->gomx, "raw"));
			break;
		case ARG_PARAMETRIC_STEREO:
			g_value_set_uint (value, g_omx_core_get_uint_param (self->gomx, "parametric_stereo"));
			break;
		case ARG_BIT_OUTPUT:
			g_value_set_uint (value, g_omx_core_get_uint_param (self->gomx, "bit_output"));
			break;
		case ARG_MIME:
			g_value_set_uint (value, g_omx_core_get_uint_param (self->gomx, "mime"));
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

/* GstElement */

#if 0
static gboolean
gst_gomx_audiodec_src_setcaps (GstPad *pad,
							   GstCaps *caps)
{
	GstGOmxAudioDec *self;
	GOmxCore *gomx;

	g_return_val_if_fail (gst_caps_get_size (caps) == 1, FALSE);

	self = GST_GOMX_AUDIODEC (GST_PAD_PARENT (pad));
	gomx = (GOmxCore *) self->gomx;

	return gst_pad_set_caps (pad, caps);
}
#endif

static gboolean
gst_gomx_audiodec_sink_setcaps (GstPad *pad,
								GstCaps *caps)
{
	GstStructure *structure;
	GstGOmxAudioDec *self;
	GOmxCore *gomx;

	g_return_val_if_fail (gst_caps_get_size (caps) == 1, FALSE);

	self = GST_GOMX_AUDIODEC (GST_PAD_PARENT (pad));
	gomx = (GOmxCore *) self->gomx;

	GST_DEBUG_OBJECT (self, ("sink setcaps\n"));

	structure = gst_caps_get_structure (caps, 0);

	switch (self->info->id)
	{
		case GST_ELEMENT_AAC:
		case GST_ELEMENT_MP3:
		case GST_ELEMENT_PCM:
			{
				guint sample_rate;
				guint channels;

				gst_structure_get_int (structure, "rate", &sample_rate);
				gst_structure_get_int (structure, "channels", &channels);
				g_omx_core_set_uint_param (gomx, "sample_rate", sample_rate);
				g_omx_core_set_uint_param (gomx, "channels", channels);
				break;
			}
	}

	g_omx_core_set_params (gomx);
	g_omx_core_set_port_cb (gomx, 0, got_input_cb);
	g_omx_core_set_port_cb (gomx, 1, got_output_cb);

	g_omx_core_prepare (gomx);
	g_omx_core_start (gomx);

	return gst_pad_set_caps (pad, caps);
}

static GstFlowReturn
gst_gomx_audiodec_chain (GstPad *pad,
						 GstBuffer *buf)
{
	GstGOmxAudioDec *self;
	GOmxCore *gomx;
	GstFlowReturn ret = GST_FLOW_OK;

	self = GST_GOMX_AUDIODEC (GST_OBJECT_PARENT (pad));

	gomx = self->gomx;

#if 0
	g_return_val_if_fail (gomx->ports[0]->buffer_size == GST_BUFFER_SIZE (buf), GST_FLOW_ERROR);
#endif
	
	g_omx_sem_down (self->input_ready);

	/** @todo Get rid of this memcpy. */
	memcpy (gomx->omx_buffer->pBuffer, GST_BUFFER_DATA (buf), GST_BUFFER_SIZE (buf));
	gomx->omx_buffer->nFilledLen = GST_BUFFER_SIZE (buf);

	g_omx_sem_up (self->input_filled);

	gst_buffer_unref (buf);

	return ret;
}

static gboolean
plugin_init (GstPlugin *plugin)
{
	int i = 0;

	GST_DEBUG_CATEGORY_INIT (gomx_audiodec_debug, "omx_audiodec",
							 0, "OpenMAX audio decoders");

	enc_global_plugins = g_hash_table_new (NULL, NULL);

	GTypeInfo typeinfo =
	{
		sizeof (GstGOmxAudioDecClass),
		(GBaseInitFunc) gst_gomx_audiodec_base_init,
		NULL,
		(GClassInitFunc) gst_gomx_audiodec_class_init,
		NULL,
		NULL,
		sizeof (GstGOmxAudioDec),
		0,
		(GInstanceInitFunc) gst_gomx_audiodec_init,
	};

	for (i = 0; comps[i].id; i++)
	{
		gchar *type_name;
		gchar *element_name;
		GType type;

		type_name = g_strdup_printf ("GstOmx%s", comps[i].capitalized_name);
		element_name = g_strdup_printf ("omx%s", comps[i].name);

		GstGOmxAudioDecClassParams *params;
		params = g_new0 (GstGOmxAudioDecClassParams, 1);
		params->info = &comps[i];
		
		params->sinkcaps = generate_sink_template (params->info->id);

		g_hash_table_insert (enc_global_plugins,
							 GINT_TO_POINTER (0), (gpointer) params);

		type = g_type_register_static (GST_TYPE_ELEMENT, type_name, &typeinfo, 0);
		if (!gst_element_register (plugin, element_name, GST_RANK_NONE, type))
		{
			g_free (type_name);
			g_free (element_name);
			return FALSE;
		}

		g_free (type_name);
		g_free (element_name);

		g_hash_table_insert (enc_global_plugins,
							 GINT_TO_POINTER (type), (gpointer) params);
	}

	g_hash_table_remove (enc_global_plugins, GINT_TO_POINTER (0));

	return TRUE;
}

GST_PLUGIN_DEFINE (GST_VERSION_MAJOR,
				   GST_VERSION_MINOR,
				   "omxaudiodec",
				   "OpenMAX audio decoders",
				   plugin_init,
				   VERSION,
				   "unknown",
				   "OpenMAX",
				   "http://ti.com/")
