/*
 * Copyright (C) 2006 Texas Instruments, Incorporated
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "gstomxaudioenc.h"

#include <gst/gst.h>
#include <config.h>

/* For memcpy. */
#include <string.h>

GST_DEBUG_CATEGORY_STATIC (gomx_audioenc_debug);
#define GST_CAT_DEFAULT gomx_audioenc_debug

#define DEFAULT_NUM_INPUT_BUFFERS 1
#define DEFAULT_NUM_OUTPUT_BUFFERS 1
#define DEFAULT_MIME 0

/* Amr encoder */
#define DEFAULT_AMR_BAND_MODE 5
#define DEFAULT_AMR_DTX_MODE OMX_AUDIO_AMRDTXModeOnVAD1
#define DEFAULT_AMR_ACDN_MODE FALSE
#define DEFAULT_AMR_EFR_MODE FALSE
#define DEFAULT_AMR_FRAME_MODE FALSE
#define GST_TYPE_AMRBM (gst_gomx_audioenc_amrbm_get_type())

typedef enum _GstOmxAudioEncElementId GstOmxAudioEncElementId;

enum _GstOmxAudioEncElementId
{
	GST_ELEMENT_NULL,
	GST_ELEMENT_AMRENC
};

GstGOmxAudioEncElementInfo comps[] =
{
	{
		.id = GST_ELEMENT_AMRENC,
		.name = "enc_amr",
		.omx_component = "OMX_AmrEncoder",
		.capitalized_name = "EncAmr",
		.longname = "AMR Encoder",
		.category = "Codec/Encoder/Audio"
	},
	{
		.id = GST_ELEMENT_NULL,
		.name = NULL,
		.omx_component = NULL,
		.capitalized_name = NULL,
		.longname = NULL,
		.category = NULL
   	}
};

typedef struct
{
	GstGOmxAudioEncElementInfo *info;

	GstCaps *srccaps;
	GstCaps *sinkcaps;
} GstGOmxAudioEncClassParams;

enum
{
	LAST_SIGNAL
};

enum
{
	ARG_0,
	ARG_SILENT,
	ARG_NUM_INPUT_BUFFERS,
	ARG_NUM_OUTPUT_BUFFERS,
	ARG_MIME,
	ARG_AMR_BAND_MODE,
	ARG_AMR_DTX_MODE,
	ARG_AMR_ACDN_MODE,
	ARG_AMR_EFR_MODE,
	ARG_AMR_FRAME_MODE
};

static GType
gst_gomx_audioenc_amrbm_get_type ()
{
	static GType gomx_audioenc_amrbm_type = 0;
	static GEnumValue gomx_audioenc_amrbm[] =
	{
		{0, "0", "BR122"},
		{1, "1", "BR102"},
		{2, "2", "BR795"},
		{3, "3", "BR74"},
		{4, "4", "BR67"},
		{5, "5", "BR59"},
		{6, "6", "BR515"},		
		{7, "7", "BR475"},
		{0, NULL, NULL},
	};
	if (!gomx_audioenc_amrbm_type)
	{
		gomx_audioenc_amrbm_type = g_enum_register_static ("GstGOmxAudioEncAmrBM", gomx_audioenc_amrbm);
	}
	return gomx_audioenc_amrbm_type;
}

#define GST_TYPE_AMRDTX (gst_gomx_audioenc_amrdtx_get_type())

static GType
gst_gomx_audioenc_amrdtx_get_type ()
{
	static GType gomx_audioenc_amrdtx_type = 0;
	static GEnumValue gomx_audioenc_amrdtx[] =
	{
		
		{OMX_AUDIO_AMRDTXModeOff,    "0", "DTX off"},
		{OMX_AUDIO_AMRDTXModeOnVAD1, "1", "Use Type 1 VAD"},
		{0, NULL, NULL},
	};
	if (!gomx_audioenc_amrdtx_type)
	{
		gomx_audioenc_amrdtx_type = g_enum_register_static ("GstGOmxAudioEncAmrDTX", gomx_audioenc_amrdtx);
	}
	return gomx_audioenc_amrdtx_type;
}

static GHashTable *enc_global_plugins;

static GstStaticPadTemplate src_factory =
GST_STATIC_PAD_TEMPLATE ("src",
						 GST_PAD_SRC,
						 GST_PAD_ALWAYS,
						 GST_STATIC_CAPS ("audio/AMR, "
										  "width =	 (int) 16, "
										  "depth = (int) 16, "
										  "signed = (boolean) TRUE, "
										  "endianness = (int) BYTE_ORDER, "
										  "rate = (int) 8000," "channels = (int) 1")
						);

/* A number of functon prototypes are given so we can refer to them later. */
static void gst_gomx_audioenc_class_init (GstGOmxAudioEncClass *klass);
static void gst_gomx_audioenc_base_init (GstGOmxAudioEncClass *klass);
static void gst_gomx_audioenc_init (GstGOmxAudioEnc *audioenc);
static void gst_gomx_audioenc_dispose (GObject *object);

static GstElementClass *parent_class = NULL;

static void gst_gomx_audioenc_set_property (GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec);
static void gst_gomx_audioenc_get_property (GObject *object, guint prop_id, GValue *value, GParamSpec *pspec);

/* static gboolean gst_gomx_audioenc_src_setcaps (GstPad *pad, GstCaps *caps); */
static gboolean gst_gomx_audioenc_sink_setcaps (GstPad *pad, GstCaps *caps);
static GstFlowReturn gst_gomx_audioenc_chain (GstPad *pad, GstBuffer *buf);

/* */

static void
got_input_cb (GOmxCore *core,
			  OMX_BUFFERHEADERTYPE *omx_buffer)
{
	GstGOmxAudioEnc *self;

	self = core->client_data;

	GST_DEBUG_OBJECT (self, "got_input_cb\n");

	core->omx_buffer = omx_buffer;
	g_omx_sem_up (self->input_ready);
	g_omx_sem_down (self->input_filled);
	core->omx_buffer = NULL;
}

static void
got_output_cb (GOmxCore *core,
			   OMX_BUFFERHEADERTYPE *omx_buffer)
{
	GstBuffer *new_buf;
	GstGOmxAudioEnc *self;

	self = core->client_data;

#if 0
	/* Print data */
	int i;
	int j;
	int counter;
	counter=0;
	for (j=0;j<10;j++)
	{
		for (i=0;i<15;i++)
		{
			printf ("%d\t",omx_buffer->pBuffer[counter]);
		}
		counter++;
		printf ("\n");
	}
#endif

	GST_DEBUG_OBJECT (self, "got_output_cb\n");

	new_buf = gst_buffer_new_and_alloc (omx_buffer->nFilledLen);
	/** @todo Get rid of this memcpy. */
	memcpy (GST_BUFFER_DATA (new_buf), omx_buffer->pBuffer, omx_buffer->nFilledLen);

	gst_buffer_set_caps (new_buf, GST_PAD_CAPS (self->srcpad));

	gst_pad_push (self->srcpad, new_buf);
}

/* */

static gboolean
gst_gomx_audioenc_sink_event (GstPad *pad,
							  GstEvent *event)
{
	GstGOmxAudioEnc *self;

	self = GST_GOMX_AUDIOENC (GST_OBJECT_PARENT (pad));

	switch (GST_EVENT_TYPE (event))
	{
		case GST_EVENT_EOS:

			g_omx_core_set_eos (self->gomx);

			/* We don't want input_cb to be blocked. */
			g_omx_sem_up (self->input_filled);

			break;

		default:
			break;
	}

	return gst_pad_event_default (pad, event);
}

static GstStateChangeReturn
gst_gomx_audioenc_change_state (GstElement *element,
								GstStateChange transition)
{
	GstStateChangeReturn ret = GST_STATE_CHANGE_SUCCESS;
	GstGOmxAudioEnc *self = GST_GOMX_AUDIOENC (element);

	switch (transition)
	{
		case GST_STATE_CHANGE_NULL_TO_READY:
			g_omx_core_init (self->gomx, self->info->omx_component);
			break;

		default:
			break;
	}

	ret = GST_ELEMENT_CLASS (parent_class)->change_state (element, transition);

	if (ret == GST_STATE_CHANGE_FAILURE)
		return ret;

	switch (transition)
	{
		case GST_STATE_CHANGE_PAUSED_TO_READY:
			g_omx_core_finish (self->gomx);
			break;

		case GST_STATE_CHANGE_READY_TO_NULL:
			g_omx_core_deinit (self->gomx);
			break;

		default:
			break;
	}

	return ret;
}

static void
gst_gomx_audioenc_dispose (GObject *obj)
{
	GstGOmxAudioEnc *self = (GstGOmxAudioEnc *) obj;

	g_omx_core_free (self->gomx);
	g_hash_table_destroy (enc_global_plugins);

	G_OBJECT_CLASS (parent_class)->dispose (obj);
}

/* */

static GstCaps *
generate_sink_template (GstOmxAudioEncElementId id)
{
	GstCaps *caps = NULL;

	if (id == GST_ELEMENT_AMRENC)
	{
		caps = gst_caps_new_simple ("audio/x-raw-int",
									"rate", G_TYPE_INT, 8000,
									"channels", G_TYPE_INT, 1,
									NULL);
	}

	return caps;
}

static void
gst_gomx_audioenc_base_init (GstGOmxAudioEncClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
	GstElementClass *element_class = GST_ELEMENT_CLASS (klass);
	GstGOmxAudioEncClassParams *params;
	GstElementDetails details;

	params = g_hash_table_lookup (enc_global_plugins,
								GINT_TO_POINTER (G_OBJECT_CLASS_TYPE (gobject_class)));

	if (!params)
	{
		params = g_hash_table_lookup (enc_global_plugins, GINT_TO_POINTER (0));
	}

	g_assert (params);

	details.longname = g_strdup_printf ("OpenMAX %s encoder",
										params->info->longname);
	details.klass = params->info->category;
	details.description = g_strdup_printf ("Encodes %s audio with OpenMAX",
										   params->info->longname);
	details.author = "Texas Instruments";
	gst_element_class_set_details (element_class, &details);
	g_free (details.longname);
	g_free (details.description);

	{
		GstPadTemplate *sinktempl;

		sinktempl = gst_pad_template_new ("sink", GST_PAD_SINK,
										  GST_PAD_ALWAYS, params->sinkcaps);

		gst_element_class_add_pad_template (element_class, sinktempl);
	}

	gst_element_class_add_pad_template (element_class,
										gst_static_pad_template_get (&src_factory));

	klass->info = params->info;
}

static void
gst_gomx_audioenc_class_init (GstGOmxAudioEncClass *klass)
{
	GObjectClass *gobject_class;
	GstElementClass *gstelement_class;

	/* gobject_class = G_OBJECT_CLASS (klass); */
	gobject_class = (GObjectClass *) klass;
	gstelement_class = (GstElementClass *) klass;

	parent_class = g_type_class_ref (GST_TYPE_ELEMENT);

	gobject_class->set_property = gst_gomx_audioenc_set_property;
	gobject_class->get_property = gst_gomx_audioenc_get_property;

	g_object_class_install_property (gobject_class, ARG_SILENT,
									 g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
														   FALSE, G_PARAM_READWRITE));

	g_object_class_install_property (gobject_class, ARG_NUM_INPUT_BUFFERS,
									 g_param_spec_uint ("input-buffers", "Input buffers",
														"The number of input buffers",
														0, G_MAXUINT, DEFAULT_NUM_INPUT_BUFFERS, G_PARAM_READWRITE));

	g_object_class_install_property (gobject_class, ARG_NUM_OUTPUT_BUFFERS,
									 g_param_spec_uint ("output-buffers", "Output buffers",
														"The number of output buffers",
														0, G_MAXUINT, DEFAULT_NUM_OUTPUT_BUFFERS, G_PARAM_READWRITE));

	if (klass->info->id == GST_ELEMENT_AMRENC)
	{
		g_object_class_install_property (gobject_class, ARG_MIME,
										 g_param_spec_uint ("mime", "MIME","Enable using mime headers",
															0, 1, DEFAULT_MIME, G_PARAM_READWRITE));

		g_object_class_install_property (gobject_class, ARG_AMR_BAND_MODE,
										 g_param_spec_enum ("band-mode", "Band Mode",
															"Specifies the encoding bit rate setting",
															GST_TYPE_AMRBM, DEFAULT_AMR_BAND_MODE, G_PARAM_READWRITE));									

		g_object_class_install_property (gobject_class, ARG_AMR_DTX_MODE,
										 g_param_spec_enum ("dtx-mode", "DTX Mode",
															"Specifies the encoding DTX settings",
															GST_TYPE_AMRDTX, DEFAULT_AMR_DTX_MODE, G_PARAM_READWRITE));	

		g_object_class_install_property (gobject_class, ARG_AMR_ACDN_MODE,
										 g_param_spec_boolean ("acdn-mode", "ACDN Mode",
															   "Specifies the encoding acoustic mode",
															   FALSE, G_PARAM_READWRITE));	

		g_object_class_install_property (gobject_class, ARG_AMR_EFR_MODE,
										 g_param_spec_boolean ("efr-mode", "EFR Mode",
															   "Specifies the encoding EFR mode",
															   FALSE, G_PARAM_READWRITE));

		g_object_class_install_property (gobject_class, ARG_AMR_FRAME_MODE,
										 g_param_spec_boolean ("frame-mode", "Frame Mode",
															   "Specifies the encoding frame mode",
															   FALSE, G_PARAM_READWRITE));
	}

	gobject_class->dispose = gst_gomx_audioenc_dispose;
	gstelement_class->change_state = gst_gomx_audioenc_change_state;
}

static void
gst_gomx_audioenc_init (GstGOmxAudioEnc *self)
{
	GstElementClass *klass = GST_ELEMENT_GET_CLASS (self);
	GstGOmxAudioEncClass *oclass = (GstGOmxAudioEncClass *) (G_OBJECT_GET_CLASS (self));

	self->sinkpad =
		gst_pad_new_from_template (gst_element_class_get_pad_template (klass, "sink"), "sink");

	gst_pad_set_setcaps_function (self->sinkpad, gst_gomx_audioenc_sink_setcaps);
	/* gst_pad_set_setcaps_function (self->sinkpad, gst_pad_proxy_setcaps); */
	gst_pad_set_chain_function (self->sinkpad, gst_gomx_audioenc_chain);
	gst_pad_set_event_function (self->sinkpad, gst_gomx_audioenc_sink_event);
	gst_element_add_pad (GST_ELEMENT (self), self->sinkpad);

	self->srcpad =
		gst_pad_new_from_template (gst_element_class_get_pad_template (klass, "src"), "src");

	/* gst_pad_set_setcaps_function (self->srcpad, gst_gomx_audioenc_src_setcaps); */
	/* gst_pad_set_setcaps_function (self->srcpad, gst_pad_proxy_setcaps); */
	/* gst_pad_use_fixed_caps (self->srcpad); */
	gst_pad_use_fixed_caps (self->sinkpad);
	gst_element_add_pad (GST_ELEMENT (self), self->srcpad);

	self->silent = FALSE;
	self->info = oclass->info;

	{
		GOmxCore *gomx;
		self->gomx = gomx = g_omx_core_new ();
		gomx->client_data = self;
		g_omx_core_set_id (gomx, self->info->name);

		g_omx_core_set_uint_param (gomx, "band_mode", DEFAULT_AMR_BAND_MODE);
		g_omx_core_set_uint_param (gomx, "dtx_mode", DEFAULT_AMR_DTX_MODE);
		g_omx_core_set_uint_param (gomx, "acdn_mode", DEFAULT_AMR_ACDN_MODE);
		g_omx_core_set_uint_param (gomx, "mime", 0);
		g_omx_core_set_uint_param (gomx, "efr_mode", DEFAULT_AMR_EFR_MODE);
		g_omx_core_set_uint_param (gomx, "frame_mode", DEFAULT_AMR_FRAME_MODE);

		gomx->num_input_buffers = DEFAULT_NUM_INPUT_BUFFERS;
		gomx->num_output_buffers = DEFAULT_NUM_OUTPUT_BUFFERS;
	}

	self->input_ready = g_omx_sem_new ();
	self->input_filled = g_omx_sem_new ();

	g_omx_sem_set_name (self->input_ready, "input ready");
	g_omx_sem_set_name (self->input_filled, "input filled");
}

static void
gst_gomx_audioenc_set_property (GObject *object,
								guint prop_id,
								const GValue *value,
								GParamSpec *pspec)
{
	GstGOmxAudioEnc *self = GST_GOMX_AUDIOENC (object);

	switch (prop_id)
	{
		case ARG_SILENT:
			self->silent = g_value_get_boolean (value);
			break;
		case ARG_NUM_INPUT_BUFFERS:
			self->gomx->num_input_buffers = g_value_get_uint (value);
			break;
		case ARG_NUM_OUTPUT_BUFFERS:
			self->gomx->num_output_buffers = g_value_get_uint (value);
			break;
		case ARG_MIME:
			g_omx_core_set_uint_param (self->gomx, "mime", g_value_get_uint (value));
			break;
		case ARG_AMR_BAND_MODE:
			g_omx_core_set_uint_param (self->gomx, "band_mode", g_value_get_enum (value));
			break;
		case ARG_AMR_DTX_MODE:
			g_omx_core_set_uint_param (self->gomx, "dtx_mode", g_value_get_enum (value));
			break;
		case ARG_AMR_ACDN_MODE:
			g_omx_core_set_uint_param (self->gomx, "acdn_mode", g_value_get_boolean (value));
			break;
		case ARG_AMR_EFR_MODE:
			g_omx_core_set_uint_param (self->gomx, "efr_mode", g_value_get_boolean (value));
			break;
		case ARG_AMR_FRAME_MODE:
			g_omx_core_set_uint_param (self->gomx, "frame_mode", g_value_get_boolean (value));
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
gst_gomx_audioenc_get_property (GObject *object,
								guint prop_id,
								GValue *value,
								GParamSpec *pspec)
{
	GstGOmxAudioEnc *self = GST_GOMX_AUDIOENC (object);

	switch (prop_id)
	{
		case ARG_SILENT:
			g_value_set_boolean (value, self->silent);
			break;
		case ARG_NUM_INPUT_BUFFERS:
			g_value_set_uint (value, self->gomx->num_input_buffers);
			break;
		case ARG_NUM_OUTPUT_BUFFERS:
			g_value_set_uint (value, self->gomx->num_output_buffers);
			break;
		case ARG_MIME:
			g_value_set_uint (value, g_omx_core_get_uint_param (self->gomx, "mime"));
			break;	
		case ARG_AMR_BAND_MODE:
			g_value_set_enum (value, g_omx_core_get_uint_param (self->gomx, "band_mode"));
			break;
		case ARG_AMR_DTX_MODE:
			g_value_set_enum (value, g_omx_core_get_uint_param (self->gomx, "dtx_mode"));
			break;	
		case ARG_AMR_ACDN_MODE:
			g_value_set_boolean (value, g_omx_core_get_uint_param (self->gomx, "acdn_mode"));
			break;	
		case ARG_AMR_EFR_MODE:
			g_value_set_boolean (value, g_omx_core_get_uint_param (self->gomx, "efr_mode"));
			break;
		case ARG_AMR_FRAME_MODE:
			g_value_set_boolean (value, g_omx_core_get_uint_param (self->gomx, "frame_mode"));
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

/* GstElement */

#if 0
static gboolean
gst_gomx_audioenc_src_setcaps (GstPad *pad,
							   GstCaps *caps)
{
	GstGOmxAudioEnc *self;
	GOmxCore *gomx;

	g_return_val_if_fail (gst_caps_get_size (caps) == 1, FALSE);

	self = GST_GOMX_AUDIOENC (GST_PAD_PARENT (pad));
	gomx = (GOmxCore *) self->gomx;

	return gst_pad_set_caps (pad, caps);
}
#endif

static gboolean
gst_gomx_audioenc_sink_setcaps (GstPad *pad,
								GstCaps *caps)
{
	GstGOmxAudioEnc *self;
	GOmxCore *gomx;

	GST_DEBUG_OBJECT (self, "sink setcaps\n");

	g_return_val_if_fail (gst_caps_get_size (caps) == 1, FALSE);

	self = GST_GOMX_AUDIOENC (GST_PAD_PARENT (pad));
	gomx = (GOmxCore *) self->gomx;

	/* GstStructure *structure; */
	/* structure = gst_caps_get_structure (caps, 0); */

	g_omx_core_set_params (gomx);

	g_omx_core_set_port_cb (gomx, 0, got_input_cb);
	g_omx_core_set_port_cb (gomx, 1, got_output_cb);

	g_omx_core_prepare (gomx);
	g_omx_core_start (gomx);

	return gst_pad_set_caps (pad, caps);
}

static GstFlowReturn
gst_gomx_audioenc_chain (GstPad *pad,
						 GstBuffer *buf)
{
	GstGOmxAudioEnc *self;
	GOmxCore *gomx;
	GstFlowReturn ret = GST_FLOW_OK;

	self = GST_GOMX_AUDIOENC (GST_OBJECT_PARENT (pad));

	gomx = self->gomx;

	/* g_return_val_if_fail (gomx->ports[0]->buffer_size == GST_BUFFER_SIZE (buf), GST_FLOW_ERROR); */

	g_omx_sem_down (self->input_ready);

	/** @todo Get rid of this memcpy. */
	memcpy (gomx->omx_buffer->pBuffer, GST_BUFFER_DATA (buf), GST_BUFFER_SIZE (buf));
	gomx->omx_buffer->nFilledLen = GST_BUFFER_SIZE (buf);

	g_omx_sem_up (self->input_filled);

	gst_buffer_unref (buf);

	return ret;
}

static gboolean
plugin_init (GstPlugin *plugin)
{
	int i = 0;

	GST_DEBUG_CATEGORY_INIT (gomx_audioenc_debug, "omx_audioenc",
							 0, "OpenMAX audio encoders");

	enc_global_plugins = g_hash_table_new (NULL, NULL);

	GTypeInfo typeinfo =
	{
		sizeof (GstGOmxAudioEncClass),
		(GBaseInitFunc) gst_gomx_audioenc_base_init,
		NULL,
		(GClassInitFunc) gst_gomx_audioenc_class_init,
		NULL,
		NULL,
		sizeof (GstGOmxAudioEnc),
		0,
		(GInstanceInitFunc) gst_gomx_audioenc_init,
	};

	for (i = 0; comps[i].id; i++)
	{
		gchar *type_name;
		gchar *element_name;

		GType type;

		type_name = g_strdup_printf ("GstOmxEnc%s", comps[i].capitalized_name);
		element_name = g_strdup_printf ("omx%s", comps[i].name);

		GstGOmxAudioEncClassParams *params;
		params = g_new0 (GstGOmxAudioEncClassParams, 1);
		params->info = &comps[i];
		params->sinkcaps = generate_sink_template (params->info->id);

		g_hash_table_insert (enc_global_plugins,
							 GINT_TO_POINTER (0), (gpointer) params);

		type = g_type_register_static (GST_TYPE_ELEMENT, type_name, &typeinfo, 0);
		if (!gst_element_register (plugin, element_name, GST_RANK_NONE, type))
		{
			g_free (type_name);
			return FALSE;
		}

		g_free (type_name);

		g_hash_table_insert (enc_global_plugins,
							 GINT_TO_POINTER (type), (gpointer) params);
	}

	g_hash_table_remove (enc_global_plugins, GINT_TO_POINTER (0));

	return TRUE;
}

GST_PLUGIN_DEFINE (GST_VERSION_MAJOR,
				   GST_VERSION_MINOR,
				   "omxaudioenc",
				   "OpenMAX audio encoders",
				   plugin_init,
				   VERSION,
				   "unknown",
				   "OpenMAX",
				   "http://ti.com/")
