/*
 * Copyright (C) 2006 Texas Instruments, Incorporated
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "gstomxsink.h"

#include <gst/gst.h>
#include <config.h>

#include <stdbool.h>

/* For memcpy. */
#include <string.h>

GST_DEBUG_CATEGORY_STATIC (gomx_sink_debug);
#define GST_CAT_DEFAULT gomx_sink_debug

#define DEFAULT_NUM_INPUT_BUFFERS 2
#define DEFAULT_NUM_OUTPUT_BUFFERS 0
#define BUFFER_SIZE 8 * 1024

/** @todo Investigate GstAudioBaseSink. */

typedef enum _GstOmxSinkElementId GstOmxSinkElementId;

enum _GstOmxSinkElementId
{
	GST_ELEMENT_NULL,
	GST_ELEMENT_ALSA_SINK
};

GstGOmxSinkElementInfo comps[] =
{
	{
		.id = GST_ELEMENT_ALSA_SINK,
		.omx_component = "OMX.st.alsa.alsasink",
		.name = "sink_alsa",
		.capitalized_name = "SinkAlsa",
		.longname = "ALSA",
		.category = "Sink/Audio"
	},
	{
		.id = GST_ELEMENT_NULL,
		.omx_component = NULL,
		.name = NULL,
		.capitalized_name = NULL,
		.longname = NULL,
		.category = NULL,
	}
};

enum
{
	LAST_SIGNAL
};

enum
{
	ARG_0,
	ARG_SILENT,
	ARG_NUM_INPUT_BUFFERS,
	ARG_NUM_OUTPUT_BUFFERS
};

typedef struct
{
	GstGOmxSinkElementInfo *info;

	GstCaps *srccaps;
	GstCaps *sinkcaps;
} GstGOmxSinkClassParams;

static GHashTable *enc_global_plugins;

#if (G_BYTE_ORDER == G_LITTLE_ENDIAN)
#define OMX_ALSA_SINK_FACTORY_ENDIANNESS "LITTLE_ENDIAN, BIG_ENDIAN"
#else
#define OMX_ALSA_SINK_FACTORY_ENDIANNESS "BIG_ENDIAN, LITTLE_ENDIAN"
#endif

static GstStaticPadTemplate sink_factory =
GST_STATIC_PAD_TEMPLATE (
						 "sink",
						 GST_PAD_SINK,
						 GST_PAD_ALWAYS,
						 GST_STATIC_CAPS ("audio/x-raw-int, "
										  "endianness = (int) { " OMX_ALSA_SINK_FACTORY_ENDIANNESS " }, "
										  "signed = (boolean) { TRUE, FALSE}, "
										  "width = (int) 32, "
										  "dept = (int) 32, "
										  "rate = (int) [1, MAX], "
										  "channels = (int) [ 1, 8 ]; "
										  "audio/x-raw-int, "
										  "endianness = (int) { " OMX_ALSA_SINK_FACTORY_ENDIANNESS "}, "
										  "signed = (boolean) { TRUE, FALSE }, "
										  "width = (int) 16, "
										  "depth = (int) 16, "
										  "rate = (int) [ 1, MAX ], "
										  "channels = (int) [ 1, 8 ]; "
										  "audio/x-raw-int, "
										  "signed = (boolean) { TRUE, FALSE }, "
										  "width = (int) 8, "
										  "depth = (int) 8, "
										  "rate = (int) [ 1, MAX ], "
										  "channels = (int) [ 1, 8 ]")
										  );

/* A number of functon prototypes are given so we can refer to them later. */
static void gst_gomx_sink_class_init (GstGOmxSinkClass *klass);
static void gst_gomx_sink_base_init (GstGOmxSinkClass *klass);
static void gst_gomx_sink_init (GstGOmxSink *sink);
static void gst_gomx_sink_dispose (GObject *object);

static GstBaseSinkClass *parent_class = NULL;

static void gst_gomx_sink_set_property (GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec);
static void gst_gomx_sink_get_property (GObject *object, guint prop_id, GValue *value, GParamSpec *pspec);

static gboolean gst_gomx_sink_setcaps (GstBaseSink *bsink, GstCaps *caps);
static GstFlowReturn gst_gomx_sink_render (GstBaseSink *bsink, GstBuffer *buf);

static void
got_input_cb (GOmxCore *core,
			  OMX_BUFFERHEADERTYPE *omx_buffer)
{
	GstGOmxSink *self;

	self = core->client_data;

	GST_LOG_OBJECT (self, "got_input_cb\n");

	core->omx_buffer = omx_buffer;
	g_omx_sem_up (self->input_ready);
	g_omx_sem_down (self->input_filled);
	core->omx_buffer = NULL;
}

/* */

static gboolean
gst_gomx_sink_start (GstBaseSink *bsink)
{
	GstGOmxSink *self;

	self = GST_GOMX_SINK (bsink);

	GST_DEBUG_OBJECT (self, "start\n");

	g_omx_core_init (self->gomx, self->info->omx_component);

	g_omx_port_setup (self->gomx->ports[0], self->gomx->num_input_buffers, BUFFER_SIZE);
	g_omx_core_set_port_cb (self->gomx, 0, got_input_cb);

	return true;
}

static gboolean
gst_gomx_sink_stop (GstBaseSink *bsink)
{
	GstGOmxSink *self;

	self = GST_GOMX_SINK (bsink);

	g_omx_core_finish (self->gomx);
	g_omx_core_deinit (self->gomx);

	return true;
}

static gboolean
gst_gomx_sink_event (GstBaseSink *bsink,
					 GstEvent *event)
{
	GstGOmxSink *self;

	self = GST_GOMX_SINK (bsink);

	switch (GST_EVENT_TYPE (event))
	{
		case GST_EVENT_EOS:

			GST_DEBUG_OBJECT (self, "eos");

			g_omx_core_set_eos (self->gomx);

			/* We don't want input_cb to be blocked. */
			g_omx_sem_up (self->input_filled);

			break;

		default:
			break;
	}

	return TRUE;
}

static void
gst_gomx_sink_dispose (GObject *obj)
{
	GstGOmxSink *self = (GstGOmxSink *) obj;

	GST_DEBUG_OBJECT (self, "dispose");

	if (self->dispose_has_run)
	{
		/* If dispose did already run, return. */
		return;
	}
	/* Make sure dispose does not run twice. */
	self->dispose_has_run = TRUE;

	g_omx_core_free (self->gomx);

	/** @todo This should go here? */
	g_hash_table_destroy (enc_global_plugins);

	g_omx_sem_free (self->input_filled);
	g_omx_sem_free (self->input_ready);

	G_OBJECT_CLASS (parent_class)->dispose (obj);
}

/* */

static void
gst_gomx_sink_base_init (GstGOmxSinkClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
	GstElementClass *element_class = GST_ELEMENT_CLASS (klass);
	GstGOmxSinkClassParams *params;
	GstGOmxSinkElementInfo *info;
	GstElementDetails details;

	params = g_hash_table_lookup (enc_global_plugins,
								  GINT_TO_POINTER (G_OBJECT_CLASS_TYPE (gobject_class)));

	if (!params)
	{
		params = g_hash_table_lookup (enc_global_plugins, GINT_TO_POINTER (0));
	}

	g_assert (params);

	info = params->info;

	details.longname = g_strdup_printf ("OpenMAX %s sink", info->longname);
	details.klass = info->category;
	details.description = g_strdup_printf ("Outputs stream to %s through OpenMAX",
										   info->longname);
	details.author = "Texas Instruments";
	gst_element_class_set_details (element_class, &details);
	g_free (details.longname);
	g_free (details.description);

	gst_element_class_add_pad_template (element_class,
										gst_static_pad_template_get (&sink_factory));

	klass->info = info;
}

static void
gst_gomx_sink_class_init (GstGOmxSinkClass *klass)
{
	GObjectClass *gobject_class;
	GstElementClass *gstelement_class;
	GstBaseSinkClass *gstbase_sink_class;

	// gobject_class = G_OBJECT_CLASS (klass);
	gobject_class = (GObjectClass *) klass;
	gstelement_class = (GstElementClass *) klass;
	gstbase_sink_class = (GstBaseSinkClass *) klass;
	gstbase_sink_class = GST_BASE_SINK_CLASS (klass);

	/* parent_class = g_type_class_ref (GST_TYPE_ELEMENT); */
	parent_class = (GstBaseSinkClass *) g_type_class_peek_parent (klass);

	gobject_class->set_property = gst_gomx_sink_set_property;
	gobject_class->get_property = gst_gomx_sink_get_property;

	g_object_class_install_property (gobject_class, ARG_SILENT,
									 g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
														   FALSE, G_PARAM_READWRITE));

	g_object_class_install_property (gobject_class, ARG_NUM_INPUT_BUFFERS,
									 g_param_spec_uint ("input-buffers", "Input buffers",
														"The number of input buffers",
														0, G_MAXUINT, DEFAULT_NUM_INPUT_BUFFERS, G_PARAM_READWRITE));

	g_object_class_install_property (gobject_class, ARG_NUM_OUTPUT_BUFFERS,
									 g_param_spec_uint ("output-buffers", "Output buffers",
														"The number of output buffers",
														0, G_MAXUINT, DEFAULT_NUM_OUTPUT_BUFFERS, G_PARAM_READWRITE));

	gobject_class->dispose = gst_gomx_sink_dispose;
	/* gobject_class->finalize = gst_gomx_sink_finalize; */

	gstbase_sink_class->event = gst_gomx_sink_event;
	gstbase_sink_class->start = gst_gomx_sink_start;
	gstbase_sink_class->stop = gst_gomx_sink_stop;
	gstbase_sink_class->preroll = gst_gomx_sink_render;
	gstbase_sink_class->render = gst_gomx_sink_render;
	gstbase_sink_class->set_caps = gst_gomx_sink_setcaps;
}

static void
gst_gomx_sink_init (GstGOmxSink *self)
{
	GstGOmxSinkClass *oclass = (GstGOmxSinkClass *) (G_OBJECT_GET_CLASS (self));

	GST_DEBUG_OBJECT (self, "init");

	self->silent = FALSE;
	self->info = oclass->info;

	{
		GOmxCore *gomx;

		self->gomx = gomx = g_omx_core_new ();
		gomx->client_data = self;

		g_omx_core_set_id (gomx, self->info->name);

		gomx->num_input_buffers = DEFAULT_NUM_INPUT_BUFFERS;
		gomx->num_output_buffers = DEFAULT_NUM_OUTPUT_BUFFERS;
	}

	self->input_ready = g_omx_sem_new ();
	self->input_filled = g_omx_sem_new ();

	g_omx_sem_set_name (self->input_ready, "input ready");
	g_omx_sem_set_name (self->input_filled, "input filled");
}

static void
gst_gomx_sink_set_property (GObject *object,
							guint prop_id,
							const GValue *value,
							GParamSpec *pspec)
{
	GstGOmxSink *self = GST_GOMX_SINK (object);

	switch (prop_id)
	{
		case ARG_SILENT:
			self->silent = g_value_get_boolean (value);
			break;
		case ARG_NUM_INPUT_BUFFERS:
			self->gomx->num_input_buffers = g_value_get_uint (value);
			break;
		case ARG_NUM_OUTPUT_BUFFERS:
			self->gomx->num_output_buffers = g_value_get_uint (value);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
gst_gomx_sink_get_property (GObject *object,
							guint prop_id,
							GValue *value,
							GParamSpec *pspec)
{
	GstGOmxSink *self = GST_GOMX_SINK (object);

	switch (prop_id)
	{
		case ARG_SILENT:
			g_value_set_boolean (value, self->silent);
			break;
		case ARG_NUM_INPUT_BUFFERS:
			g_value_set_uint (value, self->gomx->num_input_buffers);
			break;
		case ARG_NUM_OUTPUT_BUFFERS:
			g_value_set_uint (value, self->gomx->num_output_buffers);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

/* GstBaseSink */

#define DEBUG 1

static gboolean
gst_gomx_sink_setcaps (GstBaseSink *bsink,
					   GstCaps *caps)
{
	GstGOmxSink *self;
	GOmxCore *gomx;

	g_return_val_if_fail (gst_caps_get_size (caps) == 1, FALSE);

	self = GST_GOMX_SINK (bsink);
	gomx = (GOmxCore *) self->gomx;

	GST_DEBUG_OBJECT (self, "setcaps\n");

	{
		GstStructure *structure;
		structure = gst_caps_get_structure (caps, 0);

#if defined(DEBUG)
		{
			gchar *str;
			str = gst_structure_to_string (structure);

			GST_DEBUG_OBJECT (self, "set_caps = %s\n", str);

			g_free (str);
		}
#endif /* defined(DEBUG) */

		{
			gint channels;
			gst_structure_get_int (structure, "channels", &channels);
			g_omx_core_set_uint_param (gomx, "channels", channels);
		}

		{
			gint width;
			gst_structure_get_int (structure, "width", &width);
			g_omx_core_set_uint_param (gomx, "width", width);
		}

		{
			gint rate;
			gst_structure_get_int (structure, "rate", &rate);
			g_omx_core_set_int_param (gomx, "rate", rate);
		}

		{
			gint endianness;
			gst_structure_get_int (structure, "endianness", &endianness);
			g_omx_core_set_bool_param (gomx, "is_bigendian", (endianness == 1234) ? false : true);
		}

		{
			gboolean is_signed;
			gst_structure_get_boolean (structure, "signed", &is_signed);
			g_omx_core_set_bool_param (gomx, "is_signed", is_signed);
		}
	}

	g_omx_core_prepare (gomx);

	g_omx_core_start (self->gomx);

	return TRUE;
}

static GstFlowReturn
gst_gomx_sink_render (GstBaseSink *bsink,
					  GstBuffer *buf)
{
	GstGOmxSink *self;
	GOmxCore *gomx;
	GstFlowReturn ret = GST_FLOW_OK;

	self = GST_GOMX_SINK (bsink);

	gomx = self->gomx;

	/* The last buffer usually won't have the same size. */
	g_return_val_if_fail (gomx->ports[0]->buffer_size >= GST_BUFFER_SIZE (buf), GST_FLOW_ERROR);

	g_omx_sem_down (self->input_ready);

	/** @todo Get rid of this memcpy. */
	memcpy (gomx->omx_buffer->pBuffer, GST_BUFFER_DATA (buf), GST_BUFFER_SIZE (buf));
	gomx->omx_buffer->nFilledLen = GST_BUFFER_SIZE (buf);

	g_omx_sem_up (self->input_filled);

	return ret;
}

/* Plug-in */

static gboolean
plugin_init (GstPlugin *plugin)
{
	int i = 0;

	GST_DEBUG_CATEGORY_INIT (gomx_sink_debug, "omx_sink",
							 0, "OpenMAX sinks");

	enc_global_plugins = g_hash_table_new (NULL, NULL);

	GTypeInfo typeinfo =
	{
		sizeof (GstGOmxSinkClass),
		(GBaseInitFunc) gst_gomx_sink_base_init,
		NULL,
		(GClassInitFunc) gst_gomx_sink_class_init,
		NULL,
		NULL,
		sizeof (GstGOmxSink),
		0,
		(GInstanceInitFunc) gst_gomx_sink_init,
	};

	for (i = 0; comps[i].id; i++)
	{
		gchar *type_name;
		gchar *element_name;
		GType type;

		type_name = g_strdup_printf ("GstOmx%s", comps[i].capitalized_name);
		element_name = g_strdup_printf ("omx%s", comps[i].name);

		GstGOmxSinkClassParams *params;
		params = g_new0 (GstGOmxSinkClassParams, 1);
		params->info = &comps[i];

		g_hash_table_insert (enc_global_plugins,
							 GINT_TO_POINTER (0), (gpointer) params);

		type = g_type_register_static (GST_TYPE_BASE_SINK, type_name, &typeinfo, 0);
		if (!gst_element_register (plugin, element_name, GST_RANK_NONE, type))
		{
			g_free (type_name);
			g_free (element_name);
			return FALSE;
		}

		g_free (type_name);
		g_free (element_name);

		g_hash_table_insert (enc_global_plugins, GINT_TO_POINTER (type), (gpointer) params);
	}

	g_hash_table_remove (enc_global_plugins, GINT_TO_POINTER (0));

	return TRUE;
}

GST_PLUGIN_DEFINE (GST_VERSION_MAJOR,
				   GST_VERSION_MINOR,
				   "omxsink",
				   "OpenMAX sink",
				   plugin_init,
				   VERSION,
				   "unknown",
				   "OpenMAX",
				   "http://ti.com/")
