/*
 * Copyright (C) 2006 Texas Instruments, Incorporated
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "gstomxsrc.h"

#include <gst/gst.h>
#include <config.h>

/* For memcpy. */
#include <string.h>

#include <stdbool.h>

GST_DEBUG_CATEGORY_STATIC (gomx_src_debug);
#define GST_CAT_DEFAULT gomx_src_debug

#define DEFAULT_WIDTH 176
#define DEFAULT_HEIGHT 144
#define DEFAULT_FRAMERATE 15
#define DEFAULT_COLOR_FORMAT OMX_COLOR_FormatYUV420Planar

#define DEFAULT_NUM_INPUT_BUFFERS 1
#define DEFAULT_NUM_OUTPUT_BUFFERS 1

typedef enum _GstOmxSrcElementId GstOmxSrcElementId;

enum _GstOmxSrcElementId
{
	GST_ELEMENT_NULL,
	GST_ELEMENT_CAMERA_SRC
};

GstGOmxSrcElementInfo comps[] =
{
	{
		.id = GST_ELEMENT_CAMERA_SRC,
		.name = "src_camera",
		.omx_component = "OMX_Camera",
		.capitalized_name = "SrcCamera",
		.longname = "Camera Source",
		.category = "Source/Video"
	},
	{
		.id = GST_ELEMENT_NULL,
		.name = NULL,
		.omx_component = NULL,
		.capitalized_name = NULL,
		.longname = NULL,
		.category = NULL
	}
};

enum
{
	LAST_SIGNAL
};

enum
{
	ARG_0,
	ARG_SILENT,
	ARG_NUM_INPUT_BUFFERS,
	ARG_NUM_OUTPUT_BUFFERS
};

typedef struct
{
	GstGOmxSrcElementInfo *info;

	GstCaps *srccaps;
	GstCaps *sinkcaps;
} GstGOmxSrcClassParams;

static GHashTable *enc_global_plugins;

static GstBaseSrcClass *parent_class = NULL;

static GstFlowReturn gst_gomx_src_create (GstBaseSrc *bsrc, guint64 offset, guint length, GstBuffer **ret);

static gboolean gst_gomx_src_set_caps (GstBaseSrc *src, GstCaps *caps);

/* */

static gboolean
gst_gomx_src_start (GstBaseSrc *bsrc)
{
	GstGOmxSrc *self;
	self = GST_GOMX_SRC (bsrc);

	GST_DEBUG_OBJECT (self, "start");

	g_omx_core_init (self->gomx, self->info->omx_component);

	self->buffer_count = 0;

	return TRUE;
}

static gboolean
gst_gomx_src_stop (GstBaseSrc *bsrc)
{
	GstGOmxSrc *self;
	self = GST_GOMX_SRC (bsrc);

	GST_DEBUG_OBJECT (self, "stop");

	if (!self->gomx->stop)
	{
		g_omx_core_set_eos (self->gomx);

		/* We don't want output_cb to be blocked. */
		g_omx_sem_up (self->output_filled);

		g_omx_core_finish (self->gomx);
		g_omx_core_deinit (self->gomx);
		self->gomx->stop = TRUE;
	}

	return TRUE;
}

static void
gst_gomx_src_dispose (GObject *obj)
{
	GstGOmxSrc *self = (GstGOmxSrc *) obj;

	GST_DEBUG_OBJECT (self, "dispose");

	if (self->dispose_has_run)
	{
		/* If dispose did already run, return. */
		return;
	}
	/* Make sure dispose does not run twice. */
	self->dispose_has_run = TRUE;

	g_omx_core_free (self->gomx);

	/** @todo This should go here? */	
	g_hash_table_destroy (enc_global_plugins);

	g_omx_sem_free (self->output_filled);
	g_omx_sem_free (self->output_ready);

	G_OBJECT_CLASS (parent_class)->dispose (obj);
}

static gboolean
gst_gomx_src_check_get_range (GstBaseSrc *bsrc)
{
	return FALSE;
}

static void
gst_gomx_src_get_times (GstBaseSrc *basesrc,
						GstBuffer *buffer,
						GstClockTime *start,
						GstClockTime *end)
{
  GstGOmxSrc *src;

  src = GST_GOMX_SRC (basesrc);

#if 1
  {
	  GstClockTime timestamp = GST_BUFFER_TIMESTAMP (buffer);

	  if (GST_CLOCK_TIME_IS_VALID (timestamp))
	  {
		  GstClockTime duration = GST_BUFFER_DURATION (buffer);

		  if (GST_CLOCK_TIME_IS_VALID (duration))
		  {
			  *end = timestamp + duration;
		  }
		  *start = timestamp;
	  }
  }
#else
  *start = -1;
  *end = -1;
#endif
}

/* */

static GstCaps *
generate_src_template (gchar *id)
{
	GstCaps *caps = NULL;

	if (strcmp (id, "src_camera") == 0)
	{
		GstStructure *struc;

		caps = gst_caps_new_simple ("video/x-raw-yuv",
									"width", GST_TYPE_INT_RANGE, 16, 4096,
									"height", GST_TYPE_INT_RANGE, 16, 4096,
									"framerate", GST_TYPE_FRACTION_RANGE, 0, 1, 30, 1,
									NULL);

		struc = gst_caps_get_structure (caps, 0);

		{
			GValue list = { 0 };
			GValue val = { 0 };

			g_value_init (&list, GST_TYPE_LIST);
			g_value_init (&val, GST_TYPE_FOURCC);

			gst_value_set_fourcc (&val, GST_MAKE_FOURCC ('Y', 'U', 'Y', '2'));
			gst_value_list_append_value (&list, &val);

			gst_value_set_fourcc (&val, GST_MAKE_FOURCC ('U', 'Y', 'V', 'Y'));
			gst_value_list_append_value (&list, &val);

			gst_structure_set_value (struc, "format", &list);

			g_value_unset (&val);
			g_value_unset (&list);
		}
	}

	return caps;
}

static GstCaps *
generate_sink_template (gchar *id)
{
	GstCaps *caps = NULL;

	return caps;
}

static void
gst_gomx_src_set_property (GObject *object,
						   guint prop_id,
						   const GValue *value,
						   GParamSpec *pspec)
{
	GstGOmxSrc *self;
	self = GST_GOMX_SRC (object);

	switch (prop_id)
	{
		case ARG_SILENT:
			self->silent = g_value_get_boolean (value);
			break;
		case ARG_NUM_INPUT_BUFFERS:
			self->gomx->num_input_buffers = g_value_get_uint (value);
			break;
		case ARG_NUM_OUTPUT_BUFFERS:
			self->gomx->num_output_buffers = g_value_get_uint (value);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
gst_gomx_src_get_property (GObject *object,
						   guint prop_id,
						   GValue *value,
						   GParamSpec *pspec)
{
	GstGOmxSrc *self;
	self = GST_GOMX_SRC (object);

	switch (prop_id)
	{
		case ARG_SILENT:
			g_value_set_boolean (value, self->silent);
			break;
		case ARG_NUM_INPUT_BUFFERS:
			g_value_set_uint (value, self->gomx->num_input_buffers);
			break;
		case ARG_NUM_OUTPUT_BUFFERS:
			g_value_set_uint (value, self->gomx->num_output_buffers);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
gst_gomx_src_base_init (GstGOmxSrcClass *src_class)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS (src_class);
	GstElementClass *element_class = GST_ELEMENT_CLASS (src_class);
	GstGOmxSrcClassParams *params;
	GstGOmxSrcElementInfo *info;
	GstElementDetails details;

	params = g_hash_table_lookup (enc_global_plugins,
								  GINT_TO_POINTER (G_OBJECT_CLASS_TYPE (gobject_class)));

	if (!params)
	{
		params = g_hash_table_lookup (enc_global_plugins, GINT_TO_POINTER (0));
	}

	g_assert (params);

	info = params->info;

	details.longname = g_strdup_printf ("OpenMAX %s",
										info->longname);
	details.klass = info->category;
	details.description = g_strdup_printf ("Captures stream from %s through OpenMAX",
										   info->longname);
	details.author = "Texas Instruments";
	gst_element_class_set_details (element_class, &details);
	g_free (details.longname);
	g_free (details.description);

	{
		GstPadTemplate *srctempl;

		srctempl = gst_pad_template_new ("src", GST_PAD_SRC,
										 GST_PAD_ALWAYS, params->srccaps);

		gst_element_class_add_pad_template (element_class, srctempl);
	}

	src_class->info = info;
}

static void
gst_gomx_src_class_init (GstGOmxSrcClass *klass)
{
	GObjectClass *gobject_class;
	GstElementClass *gstelement_class;
	GstBaseSrcClass *gstbase_src_class;

	/* gobject_class = G_OBJECT_CLASS (klass); */
	gobject_class = (GObjectClass *) klass;
	gstelement_class = (GstElementClass *) klass;
	gstbase_src_class = (GstBaseSrcClass *) klass;

	/* parent_class = g_type_class_ref (GST_TYPE_ELEMENT); */
	parent_class = (GstBaseSrcClass *) g_type_class_peek_parent (klass);

	gobject_class->set_property = gst_gomx_src_set_property;
	gobject_class->get_property = gst_gomx_src_get_property;

	g_object_class_install_property (gobject_class, ARG_SILENT,
									 g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
														   FALSE, G_PARAM_READWRITE));

	g_object_class_install_property (gobject_class, ARG_NUM_INPUT_BUFFERS,
									 g_param_spec_uint ("input-buffers", "Input buffers",
														"The number of input buffers",
														0, G_MAXUINT, DEFAULT_NUM_INPUT_BUFFERS, G_PARAM_READWRITE));

	g_object_class_install_property (gobject_class, ARG_NUM_OUTPUT_BUFFERS,
									 g_param_spec_uint ("output-buffers", "Output buffers",
														"The number of output buffers",
														0, G_MAXUINT, DEFAULT_NUM_OUTPUT_BUFFERS, G_PARAM_READWRITE));

	gobject_class->dispose = gst_gomx_src_dispose;
	/* gobject_class->finalize = gst_gomx_src_finalize; */

	gstbase_src_class->start = gst_gomx_src_start;
	gstbase_src_class->stop = gst_gomx_src_stop;
	gstbase_src_class->create = gst_gomx_src_create;
	gstbase_src_class->check_get_range = gst_gomx_src_check_get_range;
	gstbase_src_class->set_caps = gst_gomx_src_set_caps;
	gstbase_src_class->get_times = gst_gomx_src_get_times;
}

static void
gst_gomx_src_init (GstGOmxSrc *self)
{
	GstGOmxSrcClass *oclass = (GstGOmxSrcClass *) (G_OBJECT_GET_CLASS (self));

	GST_DEBUG_OBJECT (self, "init");

	self->silent = FALSE;
	self->info = oclass->info;

	{
		GOmxCore *gomx;

		self->gomx = gomx = g_omx_core_new ();
		gomx->client_data = self;

		g_omx_core_set_id (gomx, self->info->name);

		if (self->info->id == GST_ELEMENT_CAMERA_SRC)
		{
			gomx->num_output_buffers = 1;

			/** @todo Remove this. */
			gomx->sync_output = TRUE;
		}
	}

	self->output_ready = g_omx_sem_new ();
	self->output_filled = g_omx_sem_new ();

	g_omx_sem_set_name (self->output_ready, "output ready");
	g_omx_sem_set_name (self->output_filled, "output filled");
}

/* GstElement */

static OMX_BUFFERHEADERTYPE *current_omx_buffer = NULL;

static void
got_output_cb (GOmxCore *core,
			   OMX_BUFFERHEADERTYPE *omx_buffer)
{
	GstGOmxSrc *self;

	self = (GstGOmxSrc *) core->client_data;

	GST_LOG_OBJECT (self, "got_output_cb\n");

	current_omx_buffer = omx_buffer;

	g_omx_sem_up (self->output_ready);

	g_omx_sem_down (self->output_filled);

	current_omx_buffer = NULL;
}

static gboolean
gst_gomx_src_camera_set_caps (GstGOmxSrc *self,
							  GstCaps *caps)
{
	GOmxCore *gomx;
	GstStructure *structure;
	const GValue *framerate;

	/** @todo Set the caps accordingly to the defaults. */

	guint width = DEFAULT_WIDTH; 
	guint height = DEFAULT_HEIGHT;
	guint framerate_int = DEFAULT_FRAMERATE;
	guint color_format = DEFAULT_COLOR_FORMAT;

	structure = gst_caps_get_structure (caps, 0);

#if defined(DEBUG)
	{
		gchar *str;
		str = gst_structure_to_string (structure);

		GST_DEBUG_OBJECT (self, "set_caps = %s\n", str);
		g_free (str);
	}
#endif /* defined(DEBUG) */

	gomx = self->gomx;

	gst_structure_get_int (structure, "width", &width);
	gst_structure_get_int (structure, "height", &height);
	framerate = gst_structure_get_value (structure, "framerate");

	if (width == 176 && height == 144)
	{
		/* QCIF. */
	}
	else if (width == 352 && height == 288)
	{
		/* CIF. */
	}
	else
	{
		GST_ERROR_OBJECT (self, "bad resolution %dx%d\n",
						  width, height);

		return FALSE;
	}

	if (strcmp (gst_structure_get_name (structure), "video/x-raw-yuv") == 0)
	{
		guint32 fourcc;

		if (gst_structure_get_fourcc (structure, "format", &fourcc))
		{
			switch (fourcc)
			{
				case GST_MAKE_FOURCC ('Y', 'U', 'Y', '2'):
					color_format = OMX_COLOR_FormatYCbYCr;
					break;
				case GST_MAKE_FOURCC ('U', 'Y', 'V', 'Y'):
					color_format = OMX_COLOR_FormatCbYCrY;
					break;
			}
		}
	}

	{
		gint out_buffer_size = 0;

		switch (color_format)
		{
			case OMX_COLOR_FormatYCbYCr:
			case OMX_COLOR_FormatCbYCrY:
				out_buffer_size = (width * height * 2);
				break;
#if 0
			case OMX_COLOR_Format16bitRGB565:
				out_buffer_size = (width * height * 1.5);
				break;
#endif
			default:
				break;
		}

		g_omx_port_setup (gomx->ports[0], gomx->num_output_buffers, out_buffer_size);
		g_omx_core_set_port_cb (gomx, 0, got_output_cb);
	}

	if (framerate != NULL)
	{
		self->rate_numerator = gst_value_get_fraction_numerator (framerate);
		self->rate_denominator = gst_value_get_fraction_denominator (framerate);
		framerate_int = (gst_value_get_fraction_numerator (framerate) / gst_value_get_fraction_denominator (framerate));
	}

	g_omx_core_set_uint_param (gomx, "width", width);
	g_omx_core_set_uint_param (gomx, "height", height);
	g_omx_core_set_uint_param (gomx, "framerate", framerate_int);
	g_omx_core_set_uint_param (gomx, "color_format", color_format);

	/* self->saved_caps = gst_caps_copy (caps); */

	g_omx_core_set_params (gomx);
	g_omx_core_prepare (gomx);
	g_omx_core_start (gomx);

	return TRUE;
}

static gboolean
gst_gomx_src_set_caps (GstBaseSrc *src,
					   GstCaps *caps)
{
	GstGOmxSrc *self;
	gboolean r = FALSE;

	self = GST_GOMX_SRC (src);

	GST_DEBUG_OBJECT (self, "set_caps");

	if (strcmp (self->gomx->id, "src_camera") == 0)
	{
		r = gst_gomx_src_camera_set_caps (self, caps);
	}

	return r;
}

static GstFlowReturn
gst_gomx_src_create (GstBaseSrc *bsrc,
					 guint64 offset,
					 guint length,
					 GstBuffer **ret)
{
	GstGOmxSrc *self;
	GstBuffer *buf;

	self = GST_GOMX_SRC (bsrc);

	GST_DEBUG_OBJECT (self, "create");

	{
		GOmxCore *gomx;
		gomx = self->gomx;

		g_omx_sem_down (self->output_ready);

		/* Copy the buffer. */
		{
			GstPad *srcpad;
			gint size;
			size = current_omx_buffer->nFilledLen;

			srcpad = gst_element_get_pad (GST_ELEMENT (self), "src");

			gst_pad_alloc_buffer (srcpad,
								  GST_BUFFER_OFFSET_NONE,
								  size,
								  GST_PAD_CAPS (srcpad),
								  &buf);

			/* buf = gst_buffer_new_and_alloc (size); */

			/* gst_buffer_set_caps (buf, self->saved_caps); */

			/** @todo Get rid of this memcpy. */
			memcpy (GST_BUFFER_DATA (buf), current_omx_buffer->pBuffer, size);
		}

		/* Set the timestamp. */
		{
			GstClock *clock;
			GstClockTime time;

			clock = gst_element_get_clock (GST_ELEMENT (self));

			if (clock)
			{
				time = gst_clock_get_time (clock);
				time -= gst_element_get_base_time (GST_ELEMENT (self));
				gst_object_unref (clock);
			}
			else
			{
				time = GST_CLOCK_TIME_NONE;
			}

			GST_BUFFER_TIMESTAMP (buf) = time;
			GST_BUFFER_DURATION (buf) = gst_util_uint64_scale_int (GST_SECOND, self->rate_denominator, self->rate_numerator);
		}

		GST_BUFFER_OFFSET (buf) = self->buffer_count;
		self->buffer_count++;
		GST_BUFFER_OFFSET_END (buf) = self->buffer_count;

		g_omx_sem_up (self->output_filled);
	}

	*ret = buf;

	return GST_FLOW_OK;
}

static gboolean
plugin_init (GstPlugin *plugin)
{
	int i = 0;

	GST_DEBUG_CATEGORY_INIT (gomx_src_debug, "omx_src", 0, "OpenMAX sources");

	enc_global_plugins = g_hash_table_new (NULL, NULL);

	GTypeInfo typeinfo =
	{
		sizeof (GstGOmxSrcClass),
		(GBaseInitFunc) gst_gomx_src_base_init,
		NULL,
		(GClassInitFunc) gst_gomx_src_class_init,
		NULL,
		NULL,
		sizeof (GstGOmxSrc),
		0,
		(GInstanceInitFunc) gst_gomx_src_init,
	};

	for (i = 0; comps[i].id; i++)
	{
		gchar *type_name;
		gchar *element_name;
		GType type;

		type_name = g_strdup_printf ("GstOmx%s", comps[i].capitalized_name);
		element_name = g_strdup_printf ("omx%s", comps[i].name);

		GstGOmxSrcClassParams *params;
		params = g_new0 (GstGOmxSrcClassParams, 1);
		params->info = &comps[i];

		params->srccaps = generate_src_template (params->info->name);
		params->sinkcaps = generate_sink_template (params->info->name);

		g_hash_table_insert (enc_global_plugins,
							 GINT_TO_POINTER (0), (gpointer) params);

		type = g_type_register_static (GST_TYPE_BASE_SRC, type_name, &typeinfo, 0);
		if (!gst_element_register (plugin, element_name, GST_RANK_NONE, type))
		{
			g_free (type_name);
			g_free (element_name);
			return FALSE;
		}

		g_free (type_name);
		g_free (element_name);

		g_hash_table_insert (enc_global_plugins, GINT_TO_POINTER (type), (gpointer) params);
	}

	g_hash_table_remove (enc_global_plugins, GINT_TO_POINTER (0));

	return TRUE;
}

GST_PLUGIN_DEFINE (GST_VERSION_MAJOR,
				   GST_VERSION_MINOR,
				   "omxsrc",
				   "OpenMAX src",
				   plugin_init,
				   VERSION,
				   "unknown",
				   "OpenMAX",
				   "http://ti.com/")
