/*
 * Copyright (C) 2006 Texas Instruments, Incorporated
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "gstomxvideoenc.h"

#include <gst/gst.h>
#include <config.h>

/* For memcpy. */
#include <string.h>

#include <stdbool.h>

GST_DEBUG_CATEGORY_STATIC (gomx_videoenc_debug);
#define GST_CAT_DEFAULT gomx_videoenc_debug

#define DEFAULT_WIDTH 176
#define DEFAULT_HEIGHT 144
#define DEFAULT_FRAMERATE 15
#define DEFAULT_COLOR_FORMAT OMX_COLOR_FormatYUV420Planar
#define DEFAULT_BITRATE 300000

#define DEFAULT_QPI 12
#define DEFAULT_CONTROL_RATE 1
#define DEFAULT_LEVEL 1
#define DEFAULT_H263_LEVEL OMX_VIDEO_H263Level10
#define DEFAULT_MPEG4V_LEVEL OMX_VIDEO_MPEG4Levell
#define DEFAULT_H264_LEVEL OMX_VIDEO_AVCLevel1
#define DEFAULT_GOB_INTERVAL 2

#define DEFAULT_NUM_INPUT_BUFFERS 1
#define DEFAULT_NUM_OUTPUT_BUFFERS 4
#define DEFAULT_OUTPUT_BUFFER_SIZE 26250

typedef enum _GstOmxVideoEncElementId GstOmxVideoEncElementId;

enum _GstOmxVideoEncElementId
{
	GST_ELEMENT_NULL,
	GST_ELEMENT_MPEG4,
	GST_ELEMENT_H263,
	GST_ELEMENT_H264
};

GstGOmxVideoEncElementInfo comps[] =
{
	{
		.id = GST_ELEMENT_H263,
		.name = "enc_h263",
		.omx_component = "OMX_VideoEncoder",
		.capitalized_name = "EncH263",
		.longname = "H.263 Encoder",
		.category = "Codec/Encoder/Video"
	},
	{
		.id = GST_ELEMENT_MPEG4,
		.name = "enc_mpeg4",
		.omx_component = "OMX_VideoEncoder",
		.capitalized_name = "EncMpeg4",
		.longname = "MPEG4 Encoder",
		.category = "Codec/Encoder/Video"
	},
	{
		.id = GST_ELEMENT_H264,
		.name = "enc_h264",
		.omx_component = "OMX_VideoEncoder",
		.capitalized_name = "EncH264",
		.longname = "H.264 Encoder",
		.category = "Codec/Encoder/Video"
	},
	{
		.id = GST_ELEMENT_NULL,
		.name = NULL,
		.omx_component = NULL,
		.capitalized_name = NULL,
		.longname = NULL,
		.category = NULL
   	}
};

enum
{
	LAST_SIGNAL
};

enum
{
	ARG_0,
	ARG_SILENT,
	ARG_NUM_INPUT_BUFFERS,
	ARG_NUM_OUTPUT_BUFFERS,
	ARG_BITRATE,
	ARG_QPI,
	ARG_CONTROL_RATE,
	ARG_LEVEL,
	ARG_GOB_INTERVAL,
	ARG_OUTPUT_BUFFER_SIZE
};

typedef struct
{
	GstGOmxVideoEncElementInfo *info;

	GstCaps *srccaps;
	GstCaps *sinkcaps;
} GstGOmxVideoEncClassParams;

#define GST_TYPE_CR (gst_gomx_videoenc_cr_get_type())

static GType
gst_gomx_videoenc_cr_get_type ()
{
	static GType gomx_videoenc_cr_type = 0;
	static GEnumValue gomx_videoenc_cr[] =
	{
		{OMX_Video_ControlRateDisable, "0", "disable"},
		{OMX_Video_ControlRateVariable, "1", "variable"},
		{OMX_Video_ControlRateConstant, "2", "constant"},
		{OMX_Video_ControlRateVariableSkipFrames, "3", "variableskip"},
		{OMX_Video_ControlRateConstantSkipFrames, "4", "constantskip"},
		{0, NULL, NULL},
	};
	if (!gomx_videoenc_cr_type)
	{
		gomx_videoenc_cr_type = g_enum_register_static ("GstGOmxVideoEncCr", gomx_videoenc_cr);
	}
	return gomx_videoenc_cr_type;
}

#define GST_TYPE_H263_LEVEL (gst_gomx_videoenc_h263_level_get_type())

static GType
gst_gomx_videoenc_h263_level_get_type ()
{
	static GType gomx_videoenc_h263_level_type = 0;
	static GEnumValue gomx_videoenc_h263_level[] =
	{
		{OMX_VIDEO_H263Level10, "10", "Level 10"},
		{OMX_VIDEO_H263Level20, "20", "Level 20"},
		{OMX_VIDEO_H263Level30, "30", "Level 30"},
		{OMX_VIDEO_H263Level40, "40", "Level 40"},
		{OMX_VIDEO_H263Level50, "50", "Level 50"},
		{OMX_VIDEO_H263Level60, "60", "Level 60"},
		{OMX_VIDEO_H263Level70, "70", "Level 70"},
		{0, NULL, NULL},
	};
	if (!gomx_videoenc_h263_level_type)
	{
		gomx_videoenc_h263_level_type = g_enum_register_static ("GstGOmxVideoEncH263Level", gomx_videoenc_h263_level);
	}
	return gomx_videoenc_h263_level_type;
}

#define GST_TYPE_MPEG4V_LEVEL (gst_gomx_videoenc_mpeg4v_level_get_type())

static GType
gst_gomx_videoenc_mpeg4v_level_get_type ()
{
	static GType gomx_videoenc_mpeg4v_level_type = 0;
	static GEnumValue gomx_videoenc_mpeg4v_level[] =
	{
		{OMX_VIDEO_MPEG4Levell, "1", "Level 10"},
		{OMX_VIDEO_MPEG4Level2, "2", "Level 20"},
		{OMX_VIDEO_MPEG4Level3, "3", "Level 30"},
		{OMX_VIDEO_MPEG4Level3, "4", "Level 40"},
		{0, NULL, NULL},
	};
	if (!gomx_videoenc_mpeg4v_level_type)
	{
		gomx_videoenc_mpeg4v_level_type = g_enum_register_static ("GstGOmxVideoEncMPEG4VLevel", gomx_videoenc_mpeg4v_level);
	}
	return gomx_videoenc_mpeg4v_level_type;
}

#define GST_TYPE_H264_LEVEL (gst_gomx_videoenc_h264_level_get_type())

static GType
gst_gomx_videoenc_h264_level_get_type ()
{
	static GType gomx_videoenc_h264_level_type = 0;
	static GEnumValue gomx_videoenc_h264_level[] =
	{
		{OMX_VIDEO_AVCLevel1, "1", "Level 1"},
		{OMX_VIDEO_AVCLevel1b, "1.b", "Level 1.b"},
		{OMX_VIDEO_AVCLevel11, "1.1", "Level 1.1"},
		{OMX_VIDEO_AVCLevel12, "1.2", "Level 1.2"},
		{OMX_VIDEO_AVCLevel13, "1.3", "Level 1.3"},
		{OMX_VIDEO_AVCLevel2, "2", "Level 2"},
		{OMX_VIDEO_AVCLevel21, "2.1", "Level 2.1"},
		{OMX_VIDEO_AVCLevel22, "2.2", "Level 2.2"},
		{OMX_VIDEO_AVCLevel3, "3", "Level 3"},
		{OMX_VIDEO_AVCLevel31, "3.1", "Level 3.1"},
		{OMX_VIDEO_AVCLevel32, "3.2", "Level 3.2"},
		{OMX_VIDEO_AVCLevel4, "4", "Level 4"},
		{OMX_VIDEO_AVCLevel41, "4.1", "Level 4.1"},
		{OMX_VIDEO_AVCLevel42, "4.2", "Level 4.2"},
		{OMX_VIDEO_AVCLevel5, "5", "Level 5"},
		{OMX_VIDEO_AVCLevel51, "5.1", "Level 5.1"},
		{0, NULL, NULL},
	};
	if (!gomx_videoenc_h264_level_type)
	{
		gomx_videoenc_h264_level_type = g_enum_register_static ("GstGOmxVideoEncH264Level", gomx_videoenc_h264_level);
	}
	return gomx_videoenc_h264_level_type;
}

static GHashTable *enc_global_plugins;

static GstStaticPadTemplate sink_factory =
GST_STATIC_PAD_TEMPLATE ("sink",
						 GST_PAD_SINK,
						 GST_PAD_ALWAYS,
						 GST_STATIC_CAPS ("video/x-raw-yuv, "
										  "format = (fourcc) { I420, YUY2 }, "
										  "width = (int) [16, 4096], "
										  "height = (int) [16, 4096], "
										  "framerate = (GstFraction) [0/1, 30/1]")
						);

/* A number of functon prototypes are given so we can refer to them later. */
static void gst_gomx_videoenc_class_init (GstGOmxVideoEncClass *klass);
static void gst_gomx_videoenc_base_init (GstGOmxVideoEncClass *klass);
static void gst_gomx_videoenc_init (GstGOmxVideoEnc *videoenc);
static void gst_gomx_videoenc_dispose (GObject *object);

static GstElementClass *parent_class = NULL;

static void gst_gomx_videoenc_set_property (GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec);
static void gst_gomx_videoenc_get_property (GObject *object, guint prop_id, GValue *value, GParamSpec *pspec);

/* static gboolean gst_gomx_videoenc_src_setcaps (GstPad *pad, GstCaps *caps); */
static gboolean gst_gomx_videoenc_sink_setcaps (GstPad *pad, GstCaps *caps);
static GstFlowReturn gst_gomx_videoenc_chain (GstPad *pad, GstBuffer *buf);

/* */

static void
got_input_cb (GOmxCore *core,
			  OMX_BUFFERHEADERTYPE *omx_buffer)
{
	GstGOmxVideoEnc *self;

	self = core->client_data;

	GST_LOG_OBJECT (self, "got_input_cb\n");

	core->omx_buffer = omx_buffer;
	g_omx_sem_up (self->input_ready);
	g_omx_sem_down (self->input_filled);
	core->omx_buffer = NULL;
}

static void
got_output_cb (GOmxCore *core,
			   OMX_BUFFERHEADERTYPE *omx_buffer)
{
	GstBuffer *new_buf;
	GstGOmxVideoEnc *self;

	self = core->client_data;

	GST_LOG_OBJECT (self, "got_output_cb\n");

	new_buf = gst_buffer_new_and_alloc (omx_buffer->nFilledLen);
	/** @todo Get rid of this memcpy. */
	memcpy (GST_BUFFER_DATA (new_buf), omx_buffer->pBuffer, omx_buffer->nFilledLen);

	gst_buffer_set_caps (new_buf, GST_PAD_CAPS (self->srcpad));

	gst_pad_push (self->srcpad, new_buf);
}

/* */

static gboolean
gst_gomx_videoenc_event (GstPad *pad,
						 GstEvent *event)
{
	GstGOmxVideoEnc *self;

	self = GST_GOMX_VIDEOENC (GST_OBJECT_PARENT (pad));

	switch (GST_EVENT_TYPE (event))
	{
		case GST_EVENT_EOS:

			GST_DEBUG_OBJECT (self, "eos");

			g_omx_core_set_eos (self->gomx);

			/* We don't want output_cb to be blocked. */
			g_omx_sem_up (self->input_filled);

			break;

		default:
			break;
	}

	return gst_pad_event_default (pad, event);
}

static GstStateChangeReturn
gst_gomx_videoenc_change_state (GstElement *element,
								GstStateChange transition)
{
	GstStateChangeReturn ret = GST_STATE_CHANGE_SUCCESS;

	GstGOmxVideoEnc *self = GST_GOMX_VIDEOENC (element);

	switch (transition)
	{
		case GST_STATE_CHANGE_NULL_TO_READY:
			g_omx_core_init (self->gomx, self->info->omx_component);
			break;

		default:
			break;
	}

	ret = GST_ELEMENT_CLASS (parent_class)->change_state (element, transition);

	if (ret == GST_STATE_CHANGE_FAILURE)
		return ret;

	switch (transition)
	{
		case GST_STATE_CHANGE_PAUSED_TO_READY:
			/** @todo Check this issue. */
			/* g_omx_core_finish (self->gomx); */
			break;

		case GST_STATE_CHANGE_READY_TO_NULL:
			g_omx_core_deinit (self->gomx);
			break;

		default:
			break;
	}

	return ret;
}

static void
gst_gomx_videoenc_dispose (GObject *obj)
{
	GstGOmxVideoEnc *self = (GstGOmxVideoEnc *) obj;

	if (self->dispose_has_run)
	{
		/* If dispose did already run, return. */
		return;
	}
	/* Make sure dispose does not run twice. */
	self->dispose_has_run = TRUE;

	g_omx_core_free (self->gomx);
	g_hash_table_destroy (enc_global_plugins);

	g_omx_sem_free (self->input_filled);
	g_omx_sem_free (self->input_ready);

	G_OBJECT_CLASS (parent_class)->dispose (obj);
}

/* */

static GstCaps *
generate_src_template (GstOmxVideoEncElementId id)
{
	GstCaps *caps = NULL;

	if (id == GST_ELEMENT_H263)
	{
		caps = gst_caps_new_simple ("video/x-h263",
									"width", GST_TYPE_INT_RANGE, 16, 4096,
									"height", GST_TYPE_INT_RANGE, 16, 4096,
									"framerate", GST_TYPE_FRACTION_RANGE, 0, 1, 30, 1,
									NULL);
	}
	else if (id == GST_ELEMENT_MPEG4)
	{
		caps = gst_caps_new_simple ("video/mpeg",
									"width", GST_TYPE_INT_RANGE, 16, 4096,
									"height", GST_TYPE_INT_RANGE, 16, 4096,
									"framerate", GST_TYPE_FRACTION_RANGE, 0, 1, 30, 1,
									"mpegversion", G_TYPE_INT, 4,
									"systemstream", G_TYPE_BOOLEAN, FALSE,
									NULL);
	}
	else if (id == GST_ELEMENT_H264)
	{
		caps = gst_caps_new_simple ("video/x-h264",
									"width", GST_TYPE_INT_RANGE, 16, 4096,
									"height", GST_TYPE_INT_RANGE, 16, 4096,
									"framerate", GST_TYPE_FRACTION_RANGE, 0, 1, 30, 1,
									NULL);
	}

	return caps;
}

static void
gst_gomx_videoenc_base_init (GstGOmxVideoEncClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
	GstElementClass *element_class = GST_ELEMENT_CLASS (klass);
	GstGOmxVideoEncClassParams *params;
	GstElementDetails details;

	params = g_hash_table_lookup (enc_global_plugins,
								  GINT_TO_POINTER (G_OBJECT_CLASS_TYPE (gobject_class)));

	if (!params)
	{
		params = g_hash_table_lookup (enc_global_plugins, GINT_TO_POINTER (0));
	}

	g_assert (params);

	details.longname = g_strdup_printf ("OpenMAX %s encoder",
										params->info->longname);
	details.klass = params->info->category;
	details.description = g_strdup_printf ("Encodes %s video with OpenMAX",
										   params->info->longname);
	details.author = "Texas Instruments";
	gst_element_class_set_details (element_class, &details);
	g_free (details.longname);
	g_free (details.description);

	{
		GstPadTemplate *srctempl;

		srctempl = gst_pad_template_new ("src", GST_PAD_SRC,
										 GST_PAD_ALWAYS, params->srccaps);

		gst_element_class_add_pad_template (element_class, srctempl);
	}

	gst_element_class_add_pad_template (element_class,
										gst_static_pad_template_get (&sink_factory));

	klass->info = params->info;
}

static void
gst_gomx_videoenc_class_init (GstGOmxVideoEncClass *klass)
{
	GObjectClass *gobject_class;
	GstElementClass *gstelement_class;

	/* gobject_class = G_OBJECT_CLASS (klass); */
	gobject_class = (GObjectClass *) klass;
	gstelement_class = (GstElementClass *) klass;

	parent_class = g_type_class_ref (GST_TYPE_ELEMENT);

	gobject_class->set_property = gst_gomx_videoenc_set_property;
	gobject_class->get_property = gst_gomx_videoenc_get_property;

	g_object_class_install_property (gobject_class, ARG_SILENT,
									 g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
														   FALSE, G_PARAM_READWRITE));

	g_object_class_install_property (gobject_class, ARG_NUM_INPUT_BUFFERS,
									 g_param_spec_uint ("input-buffers", "Input buffers",
														"The number of input buffers",
														0, G_MAXUINT, DEFAULT_NUM_INPUT_BUFFERS, G_PARAM_READWRITE));

	g_object_class_install_property (gobject_class, ARG_NUM_OUTPUT_BUFFERS,
									 g_param_spec_uint ("output-buffers", "Output buffers",
														"The number of output buffers",
														0, G_MAXUINT, DEFAULT_NUM_OUTPUT_BUFFERS, G_PARAM_READWRITE));

	if (klass->info->id == GST_ELEMENT_H263 ||
		klass->info->id == GST_ELEMENT_MPEG4 ||
		klass->info->id == GST_ELEMENT_H264)
	{
		g_object_class_install_property (gobject_class, ARG_BITRATE,
										 g_param_spec_ulong ("bitrate", "Bitrate", "The Bitrate",
															 0, G_MAXULONG, DEFAULT_BITRATE, G_PARAM_READWRITE));

		g_object_class_install_property (gobject_class, ARG_CONTROL_RATE,
										 g_param_spec_enum ("controlrate", "Control Rate",
															"Specifies the control rate type",
															GST_TYPE_CR, DEFAULT_CONTROL_RATE, G_PARAM_READWRITE));
	}

	if (klass->info->id == GST_ELEMENT_H263)
	{
		g_object_class_install_property (gobject_class, ARG_QPI,
										 g_param_spec_ulong ("qpi", "QP Index",
															 "QP value to use for index frames",
															 0, G_MAXULONG, DEFAULT_QPI, G_PARAM_READWRITE));

		g_object_class_install_property (gobject_class, ARG_LEVEL,
										 g_param_spec_enum ("level", "Level",
															"Level(s) to use",
															GST_TYPE_H263_LEVEL, DEFAULT_H263_LEVEL, G_PARAM_READWRITE));
	}

	if (klass->info->id == GST_ELEMENT_MPEG4)
	{
		g_object_class_install_property (gobject_class, ARG_QPI,
										 g_param_spec_ulong ("qpi", "QP Index",
															 "QP value to use for index frames",
															 0, G_MAXULONG, DEFAULT_QPI, G_PARAM_READWRITE));

		g_object_class_install_property (gobject_class, ARG_LEVEL,
										 g_param_spec_enum ("level", "Level",
															"Level(s) to use",
															GST_TYPE_MPEG4V_LEVEL, DEFAULT_MPEG4V_LEVEL, G_PARAM_READWRITE));
	}

	if (klass->info->id == GST_ELEMENT_H264)
	{
		g_object_class_install_property (gobject_class, ARG_LEVEL,
										 g_param_spec_enum ("level", "Level",
															"Level(s) to use",
															GST_TYPE_H264_LEVEL, DEFAULT_H263_LEVEL, G_PARAM_READWRITE));

		g_object_class_install_property (gobject_class, ARG_OUTPUT_BUFFER_SIZE,
										 g_param_spec_uint ("output-buffer-size", "Output buffer size",
															"The size of the output buffers",
															0, G_MAXUINT, DEFAULT_OUTPUT_BUFFER_SIZE, G_PARAM_READWRITE));
	}

	if (klass->info->id == GST_ELEMENT_H263)
	{
		g_object_class_install_property (gobject_class, ARG_GOB_INTERVAL,
										 g_param_spec_ulong ("gobinterval", "GOB Header Interval",
															 "Specifies the interval of non-empty GOB headers in units of GOBs",
															 0, G_MAXULONG, DEFAULT_GOB_INTERVAL, G_PARAM_READWRITE));
	}

	gobject_class->dispose = gst_gomx_videoenc_dispose;
	gstelement_class->change_state = gst_gomx_videoenc_change_state;
}

static void
gst_gomx_videoenc_init (GstGOmxVideoEnc *self)
{
	GstElementClass *klass = GST_ELEMENT_GET_CLASS (self);
	GstGOmxVideoEncClass *oclass = (GstGOmxVideoEncClass *) (G_OBJECT_GET_CLASS (self));

	GST_DEBUG_OBJECT (self, "init");

	self->silent = FALSE;
	self->info = oclass->info;

	self->sinkpad =
		gst_pad_new_from_template (gst_element_class_get_pad_template (klass, "sink"), "sink");

	gst_pad_set_setcaps_function (self->sinkpad, gst_gomx_videoenc_sink_setcaps);

	self->srcpad =
		gst_pad_new_from_template (gst_element_class_get_pad_template (klass, "src"), "src");

	/* gst_pad_set_setcaps_function (self->srcpad, gst_gomx_videoenc_src_setcaps); */
	gst_pad_set_setcaps_function (self->srcpad, gst_pad_proxy_setcaps);
	gst_pad_use_fixed_caps (self->srcpad);

	gst_element_add_pad (GST_ELEMENT (self), self->sinkpad);
	gst_element_add_pad (GST_ELEMENT (self), self->srcpad);
	gst_pad_set_chain_function (self->sinkpad, gst_gomx_videoenc_chain);
	gst_pad_set_event_function (self->sinkpad, gst_gomx_videoenc_event);

	{
		GOmxCore *gomx;

		self->gomx = gomx = g_omx_core_new ();
		self->gomx->client_data = self;
		g_omx_core_set_id (gomx, self->info->name);

		g_omx_core_set_uint_param (gomx, "bitrate", DEFAULT_BITRATE);
		g_omx_core_set_uint_param (gomx, "qpi", DEFAULT_QPI);
		g_omx_core_set_uint_param (gomx, "control_rate", DEFAULT_CONTROL_RATE);
		g_omx_core_set_uint_param (gomx, "level", DEFAULT_LEVEL);
		g_omx_core_set_uint_param (gomx, "gob_interval", DEFAULT_GOB_INTERVAL);
		g_omx_core_set_uint_param (gomx, "output_buffer_size", DEFAULT_OUTPUT_BUFFER_SIZE);

		gomx->num_input_buffers = DEFAULT_NUM_INPUT_BUFFERS;
		gomx->num_output_buffers = DEFAULT_NUM_OUTPUT_BUFFERS;
	}

	self->input_ready = g_omx_sem_new ();
	self->input_filled = g_omx_sem_new ();

	g_omx_sem_set_name (self->input_ready, "input ready");
	g_omx_sem_set_name (self->input_filled, "input filled");
}

static void
gst_gomx_videoenc_set_property (GObject *object,
								guint prop_id,
								const GValue *value,
								GParamSpec *pspec)
{
	GstGOmxVideoEnc *self = GST_GOMX_VIDEOENC (object);

	switch (prop_id)
	{
		case ARG_SILENT:
			self->silent = g_value_get_boolean (value);
			break;
		case ARG_NUM_INPUT_BUFFERS:
			self->gomx->num_input_buffers = g_value_get_uint (value);
			break;
		case ARG_NUM_OUTPUT_BUFFERS:
			self->gomx->num_output_buffers = g_value_get_uint (value);
			break;
		case ARG_BITRATE:
			g_omx_core_set_uint_param (self->gomx, "bitrate", g_value_get_ulong (value));
			break;
		case ARG_QPI:
			g_omx_core_set_uint_param (self->gomx, "qpi", g_value_get_ulong (value));
			break;
		case ARG_CONTROL_RATE:
			g_omx_core_set_uint_param (self->gomx, "control_rate", g_value_get_enum (value));
			break;
		case ARG_LEVEL:
			g_omx_core_set_uint_param (self->gomx, "level", g_value_get_enum (value));
			break;
		case ARG_GOB_INTERVAL:
			g_omx_core_set_uint_param (self->gomx, "gob_interval", g_value_get_ulong (value));
			break;
		case ARG_OUTPUT_BUFFER_SIZE:
			g_omx_core_set_uint_param (self->gomx, "output_buffer_size", g_value_get_ulong (value));
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
gst_gomx_videoenc_get_property (GObject *object,
								guint prop_id,
								GValue *value,
								GParamSpec *pspec)
{
	GstGOmxVideoEnc *self = GST_GOMX_VIDEOENC (object);

	switch (prop_id)
	{
		case ARG_SILENT:
			g_value_set_boolean (value, self->silent);
			break;
		case ARG_NUM_INPUT_BUFFERS:
			g_value_set_uint (value, self->gomx->num_input_buffers);
			break;
		case ARG_NUM_OUTPUT_BUFFERS:
			g_value_set_uint (value, self->gomx->num_output_buffers);
			break;
		case ARG_BITRATE:
			g_value_set_ulong (value, g_omx_core_get_uint_param (self->gomx, "bitrate"));
			break;
		case ARG_QPI:
			g_value_set_ulong (value, g_omx_core_get_uint_param (self->gomx, "qpi"));
			break;
		case ARG_CONTROL_RATE:
			g_value_set_enum (value, g_omx_core_get_uint_param (self->gomx, "control_rate"));
			break;
		case ARG_LEVEL:
			g_value_set_enum (value, g_omx_core_get_uint_param (self->gomx, "level"));
			break;
		case ARG_GOB_INTERVAL:
			g_value_set_ulong (value, g_omx_core_get_uint_param (self->gomx, "gob_interval"));
			break;
		case ARG_OUTPUT_BUFFER_SIZE:
			g_value_set_uint (value, g_omx_core_get_uint_param (self->gomx, "output_buffer_size"));
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

/* GstElement */

#if 0
static gboolean
gst_gomx_videoenc_src_setcaps (GstPad *pad,
							   GstCaps *caps)
{
	GstGOmxVideoEnc *self;
	GOmxCore *gomx;

	g_return_val_if_fail (gst_caps_get_size (caps) == 1, FALSE);

	self = GST_GOMX_VIDEOENC (GST_PAD_PARENT (pad));

	gomx = (GOmxCore *) self->gomx;

#if 0
	GstStructure *structure;
	const GValue *framerate;
	structure = gst_caps_get_structure (caps, 0);

	gst_structure_get_int (structure, "width", &gomx->width);
	gst_structure_get_int (structure, "height", &gomx->height);
	framerate = gst_structure_get_value (structure, "framerate");

	if (gomx->width == 176 && gomx->height == 144)
	{
		/* QCIF. */
	}
	else if (width == 352 && height == 288)
	{
		/* CIF. */
	}
	else
	{
		GST_ERROR_OBJECT (self, "bad resolution %dx%d\n",
						  width, height);

		return FALSE;
	}

	if (framerate != NULL)
	{
		gomx->framerate = gst_value_get_fraction_numerator (framerate) / gst_value_get_fraction_denominator (framerate);
	}

	if (gomx->sinkset)
	{
		g_omx_core_set_params (gomx);
		g_omx_core_prepare (gomx);
	}
#endif

	return gst_pad_set_caps (pad, caps);
}
#endif

static gboolean
gst_gomx_videoenc_sink_setcaps (GstPad *pad,
								GstCaps *caps)
{
	GstStructure *structure;
	GstGOmxVideoEnc *self;
	GOmxCore *gomx;
	const GValue *framerate;

	/** @todo Check if all the caps are OK. */
	
	guint width = DEFAULT_WIDTH; 
	guint height = DEFAULT_HEIGHT;
	guint framerate_int = DEFAULT_FRAMERATE;
	guint color_format = DEFAULT_COLOR_FORMAT;
	guint compression_format = OMX_VIDEO_CodingUnused;

	g_return_val_if_fail (gst_caps_get_size (caps) == 1, FALSE);

	self = GST_GOMX_VIDEOENC (GST_PAD_PARENT (pad));
	gomx = (GOmxCore *) self->gomx;

	structure = gst_caps_get_structure (caps, 0);

	gst_structure_get_int (structure, "width", &width);
	gst_structure_get_int (structure, "height", &height);
	framerate = gst_structure_get_value (structure, "framerate");

	if (width == 176 && height == 144)
	{
		/* QCIF. */
	}
	else if (width == 352 && height == 288)
	{
		/* CIF. */
	}
	else
	{
		GST_ERROR_OBJECT (self, "bad resolution %dx%d\n",
						  width, height);

		return FALSE;
	}

	if (strcmp (gst_structure_get_name (structure), "video/x-raw-yuv") == 0)
	{
		guint32 fourcc;

		if (gst_structure_get_fourcc (structure, "format", &fourcc))
		{
			switch (fourcc)
			{
				case GST_MAKE_FOURCC ('I', '4', '2', '0'):
					color_format = OMX_COLOR_FormatYUV420Planar;
					break;
				case GST_MAKE_FOURCC ('Y', 'U', 'Y', '2'):
					color_format = OMX_COLOR_FormatYCbYCr;
					break;
			}
		}
	}

	if (framerate != NULL)
	{
		framerate_int = gst_value_get_fraction_numerator (framerate) / gst_value_get_fraction_denominator (framerate);
	}

	if (self->info->id == GST_ELEMENT_H263)
	{
		compression_format = OMX_VIDEO_CodingH263;
	}
	else if (self->info->id == GST_ELEMENT_MPEG4)
	{
		compression_format = OMX_VIDEO_CodingMPEG4;
	}
	else if (self->info->id == GST_ELEMENT_H264)
	{
		compression_format = OMX_VIDEO_CodingAVC;
	}

	g_omx_core_set_uint_param (gomx, "width", width);
	g_omx_core_set_uint_param (gomx, "height", height);
	g_omx_core_set_uint_param (gomx, "framerate", framerate_int);
	g_omx_core_set_uint_param (gomx, "color_format", color_format);
	g_omx_core_set_uint_param (gomx, "compression_format", compression_format);

	g_omx_core_set_params (gomx);
	g_omx_core_set_port_cb (gomx, 0, got_input_cb);
	g_omx_core_set_port_cb (gomx, 1, got_output_cb);

	g_omx_core_prepare (gomx);
	g_omx_core_start (gomx);

	return gst_pad_set_caps (pad, caps);
}

static GstFlowReturn
gst_gomx_videoenc_chain (GstPad *pad,
						 GstBuffer *buf)
{
	GstGOmxVideoEnc *self;
	GOmxCore *gomx;
	GstFlowReturn ret = GST_FLOW_OK;
	static int c = 0;

	self = GST_GOMX_VIDEOENC (GST_OBJECT_PARENT (pad));

	GST_LOG_OBJECT (self, "chain [%d]\n", c++);

	gomx = self->gomx;

#if 0
	g_return_val_if_fail (gomx->ports[0]->buffer_size == GST_BUFFER_SIZE (buf), GST_FLOW_ERROR);
#endif

	g_omx_sem_down (self->input_ready);

	if (gomx->omx_buffer)
	{
		/** @todo Get rid of this memcpy. */
		memcpy (gomx->omx_buffer->pBuffer, GST_BUFFER_DATA (buf), GST_BUFFER_SIZE (buf));
		gomx->omx_buffer->nFilledLen = GST_BUFFER_SIZE (buf);
	}
	else
	{
		GST_WARNING_OBJECT (self, "bad buffer\n");
	}

	g_omx_sem_up (self->input_filled);

	gst_buffer_unref (buf);

	return ret;
}

static gboolean
plugin_init (GstPlugin *plugin)
{
	int i = 0;

	GST_DEBUG_CATEGORY_INIT (gomx_videoenc_debug, "omx_videoenc",
							 0, "OpenMAX video encoders");

	enc_global_plugins = g_hash_table_new (NULL, NULL);

	GTypeInfo typeinfo =
	{
		sizeof (GstGOmxVideoEncClass),
		(GBaseInitFunc) gst_gomx_videoenc_base_init,
		NULL,
		(GClassInitFunc) gst_gomx_videoenc_class_init,
		NULL,
		NULL,
		sizeof (GstGOmxVideoEnc),
		0,
		(GInstanceInitFunc) gst_gomx_videoenc_init,
	};

	for (i = 0; comps[i].id; i++)
	{
		gchar *type_name;
		gchar *element_name;
		GType type;

		type_name = g_strdup_printf ("GstOmx%s", comps[i].capitalized_name);
		element_name = g_strdup_printf ("omx%s", comps[i].name);

		GstGOmxVideoEncClassParams *params;
		params = g_new0 (GstGOmxVideoEncClassParams, 1);
		params->info = &comps[i];

		params->srccaps = generate_src_template (params->info->id);

		g_hash_table_insert (enc_global_plugins,
							 GINT_TO_POINTER (0), (gpointer) params);

		type = g_type_register_static (GST_TYPE_ELEMENT, type_name, &typeinfo, 0);
		if (!gst_element_register (plugin, element_name, GST_RANK_NONE, type))
		{
			g_free (type_name);
			g_free (element_name);
			return FALSE;
		}

		g_free (type_name);
		g_free (element_name);

		g_hash_table_insert (enc_global_plugins,
							 GINT_TO_POINTER (type), (gpointer) params);
	}

	g_hash_table_remove (enc_global_plugins, GINT_TO_POINTER (0));

	return TRUE;
}

GST_PLUGIN_DEFINE (GST_VERSION_MAJOR,
				   GST_VERSION_MINOR,
				   "omxvideoenc",
				   "OpenMAX video encoders",
				   plugin_init,
				   VERSION,
				   "unknown",
				   "OpenMAX",
				   "http://ti.com/")
